
#include <stdio.h>

#include "windows.h"

#include "peexe.h"

#include "thread.h"
#include "Segment.h"
#include "bin32.h"

#include "Log.h"
#include "ctype.h"

#include "setjmp.h"
#include "LoadPE.h"
#include "BinTypes.h"
#include "Resources.h"
#include "Module.h"
#include "compat.h"

#define DEBUG_BIN32
#ifdef  DEBUG_BIN32
static int lf_header  = LF_CONSOLE;
static int lf_header2 = LF_CONSOLE; 
static int lf_imports = LF_CONSOLE;
static int lf_exports = LF_CONSOLE;
static int lf_log     = LF_CONSOLE;
#else
static int lf_header  = LF_LOG;
static int lf_header2 = LF_LOG; 
static int lf_imports = LF_LOG;
static int lf_exports = LF_LOG;
static int lf_log     = LF_LOG;
#endif

DWORD setstub(char *,int);

/*
 * Load a PE file into memory
 * read headers first...
 * get file ImageBase
 * get file SizeOfImage
 *
 * VirtualAlloc(ImageBase,SizeofImage,...)
 * read section headers
 * for each section read section...
 */

static MODULE_HEADERS *NTModules;
static IMAGE_NT_HEADERS NTHeader[256];
static int NTHeaderCount;

static MODULE_HEADERS *CurrentProcess;

void DumpHeader(PIMAGE_FILE_HEADER );
void DumpImportsSection(void *, PIMAGE_NT_HEADERS);
void DumpSectionTable(void *,PIMAGE_SECTION_HEADER , unsigned );
void DumpOptionalHeader(PIMAGE_OPTIONAL_HEADER );
int LoadImportsSection(char *,void * , PIMAGE_NT_HEADERS);
void LoadExportsTable(MODULE_HEADERS *,PIMAGE_NT_HEADERS ,char *);
MODULE_HEADERS * ExecEntryPoint(MODULE_HEADERS *, char *,char *);

MODULE_HEADERS * ExecPE(char *,char *);
HMODULE    GetModuleFromInstance(HINSTANCE);
HINSTANCE  TWIN_GetInstanceFromTask(HTASK);

extern MAPTABLE ModuleMapper_user32[];
extern MAPTABLE ModuleMapper_kernel32[];
extern MAPTABLE ModuleMapper_shell32[];
extern MAPTABLE ModuleMapper_gdi32[];

static MODULE_HEADERS LocalHeader[256];
static int            LocalCount;

static MODULE_HEADERS *
AllocModuleHeader(char *modulename,char *modulealias)
{
	MODULE_HEADERS *lpNTModule;

	lpNTModule = &LocalHeader[LocalCount++];

	memset(lpNTModule,0,sizeof(MODULE_HEADERS));

	lpNTModule->modulename = modulename;
	lpNTModule->modulealias = modulealias;
 	lpNTModule->modulenext  = NTModules;	

	NTModules = lpNTModule;

	return lpNTModule;
}

typedef struct {
	char *dllname;		/* name embedded within application */
	char *libname;		/* native shared library to load    */
	int   rtflag;		/* run-time flags, (loadable?)      */
				/*      0 	cannot be loaded    */
				/*      1 	try to load         */
	int   ldflag;		/* which to load, 		    */
				/*
 				 * 	1	library can be replaced w/ 
			         *		ms versionr, we don't have one
 				 *	2	DO NOT use ms version
 			   	 *	4	USE TWIN version, regardless
 				 *	8	library can be replaced w/ ms 
				 *		version, although we have one
 				 */
	HMODULE		 hModule;    /* HMODULE for this library */
	MODULE_HEADERS  *lpNTModule; /* pointer to LOADED binary information */
	MAPTABLE 	*maptable;   /* pointer to binary interface functions */
} LIBINITDATA;

LIBINITDATA LibInitData[] = {
	{ "user32.dll",    "user",      1,4,0,0,ModuleMapper_user32    },
	{ "kernel32.dll",  "kernel",    1,4,0,0,ModuleMapper_kernel32  },
	{ "gdi32.dll",     "gdi",    	1,4,0,0,ModuleMapper_gdi32     },
	{ "shell32.dll",   "shell",     1,4,0,0,ModuleMapper_shell32   },
	{ "comdlg32.dll",  "commdlg",   1,8,0,0,0     },
	{ "version.dll",   "ver",       1,8,0,0,0     },
	{ "lz32.dll",      "lz32",      1,8,0,0,0     },
	{ "comctl32.dll",  "comctl32",  1,8,0,0,0     },
	{ "winmm.dll",     "mmsystem",  1,8,0,0,0     },
	{ "rpcrt4.dll",     0,		0,2,0,0,0     },	
	{ "advapi32.dll",   0,		0,2,0,0,0     },
	{ "msvcrt.dll",     0, 		0,2,0,0,0     }, 
	{ "ntdll.dll",      0, 		0,2,0,0,0     },
	{ 0,0,0,0 }
};
	
MODULE_HEADERS *
ExecPE(char *lpszFileName,char *lpszBase)
{
	static int InitPE;
	static void *BaseAddress;
	IMAGE_DOS_HEADER DosHeader;
 	PIMAGE_SECTION_HEADER pSectionHeaders;
	PIMAGE_NT_HEADERS pNTHeader;
	int i,len;
	char *bp;
	int ret;
	HFILE hFile;
	MODULE_HEADERS *lpLibrary = 0;
	MODULE_HEADERS *lpNTModule;
	OFSTRUCT	ofs;
	LIBINITDATA *lpLibInitData = LibInitData;

	logstr(LF_CONSOLE,"ExecPE(%s,%s)\n",
		lpszFileName,lpszBase?lpszBase:"0"); 

	if(InitPE == 0) {
		InitPE++;
		LoadLibrary("commdlg");
	 	LoadLibrary("shell");
	 	LoadLibrary("comctl32");
	 	LoadLibrary("ver");
	}

	bp = lpszFileName;

	/* see if we have already loaded the desired module, if so return it */
	for(lpNTModule=NTModules;lpNTModule;lpNTModule=lpNTModule->modulenext) {
		if(strcasecmp(lpNTModule->modulename,lpszFileName) == 0) {
			return lpNTModule;
		}
	}
	
	/* check for one of our internal builtin flavors */
	for(lpLibInitData = LibInitData;lpLibInitData->dllname;lpLibInitData++){ 
		//LPMODULEINFO *modinfo;
		HMODULE       hModule;

		if(strcasecmp(lpLibInitData->dllname,lpszFileName) == 0) {
		
  		   if(lpLibInitData->lpNTModule)
			return lpLibInitData->lpNTModule;

		   if(lpLibInitData->rtflag == 0) {
		   	logstr(LF_CONSOLE,"Cannot load DLL: %s, no mapping table\n",
				lpszFileName);
			return 0;
		   }

		   hModule = GetModuleHandle(lpLibInitData->libname);

		   if(hModule == 0) {
			if(LoadLibrary(lpLibInitData->libname))
		          hModule = GetModuleHandle(lpLibInitData->libname);
		   }

		   logstr(LF_CONSOLE,"Load DLL: %s (%s) %x\n",
			lpLibInitData->dllname,
			lpLibInitData->libname,
			hModule);

		   lpNTModule = AllocModuleHeader( 
			lpLibInitData->dllname,
			lpLibInitData->libname);

		   if(lpLibInitData->maptable && lpNTModule)
			lpNTModule->maptable = lpLibInitData->maptable;

		   /* save the loaded module */
		   lpLibInitData->lpNTModule = lpNTModule;

		   return lpNTModule;
		}
	}

	/* none of the above, so try to open the win32 .dll */

logstr(LF_CONSOLE,"OpenFile(%s)\n",lpszFileName);

    	hFile = OpenFile(lpszFileName,&ofs,OF_READ);
	if(hFile == -1) {
		logstr(LF_CONSOLE,"cannot open file %s\n",lpszFileName);
		return 0;
	}

	/* read the dos image header first */
        ret = _lread(hFile,&DosHeader,sizeof(IMAGE_DOS_HEADER));

	if(DosHeader.e_magic == IMAGE_DOS_SIGNATURE)	{

		/* now read in the nt header */
    		_llseek(hFile,DosHeader.e_lfanew,0);

		pNTHeader = &NTHeader[NTHeaderCount++];
		ret = _lread(hFile,pNTHeader, sizeof(IMAGE_NT_HEADERS));

		/* yes, it is a win32 header */
		if (pNTHeader->Signature != IMAGE_NT_SIGNATURE) {
			_lclose(hFile);
			return 0;
		}

		bp = strrchr(lpszFileName,'/');
		if(bp)
			bp++;
		else 
			bp = lpszFileName;

		BaseAddress = VirtualAlloc(
			(void *) pNTHeader->OptionalHeader.ImageBase,
			pNTHeader->OptionalHeader.SizeOfImage,
			MEM_COMMIT,
			PAGE_EXECUTE_READWRITE);

		/* this will show the dll that we are being loaded for
 		 * as a dependency.  lpszBase is the library that requires
		 * the library lpszFileName.
 		 */
		if(lpszBase) 
		   logstr(LF_CONSOLE,"Load File: %s %p (%s)\n",
			lpszFileName,BaseAddress,lpszBase);
	        else
		   logstr(LF_CONSOLE,"Load File: %s %p\n",lpszFileName,BaseAddress);

		lpNTModule = AllocModuleHeader(bp,bp);
		lpNTModule->pNTHeader = pNTHeader;
		lpNTModule->BaseAddress = BaseAddress;

		/* show the NT header */
		DumpHeader(&pNTHeader->FileHeader);

		/* show the Optional header */
		DumpOptionalHeader((PIMAGE_OPTIONAL_HEADER) 
			&pNTHeader->OptionalHeader);

		pSectionHeaders = (PIMAGE_SECTION_HEADER)
			((void *)BaseAddress + sizeof(IMAGE_NT_HEADERS));

		/* now read the section headers */
		ret = _lread( hFile, pSectionHeaders, 
			sizeof(IMAGE_SECTION_HEADER)*
			pNTHeader->FileHeader.NumberOfSections);

		for(i=0; i < pNTHeader->FileHeader.NumberOfSections; i++) {
			void *LoadAddress;

			LoadAddress = 
				RVA(BaseAddress,pSectionHeaders->VirtualAddress);
		        logstr(lf_log,"Load Section %d address=%p name=%s\n",
				i,LoadAddress,pSectionHeaders->Name);

			/*if (index == 0)  */
			{
			   DumpSectionTable( LoadAddress,pSectionHeaders,i);
			}

			/* load only non-BSS segments */
			if(!(pSectionHeaders->Characteristics &
				IMAGE_SCN_CNT_UNINITIALIZED_DATA)) 
		        {
		            logstr(lf_log,"Read Section %d\n",i);

			    _llseek(hFile,pSectionHeaders->PointerToRawData,SEEK_SET);
			    len = _lread(hFile,(char*) LoadAddress, 
				pSectionHeaders->SizeOfRawData);

			    if( len != pSectionHeaders->SizeOfRawData)
			    {
				logstr(LF_ERROR,"Failed to load section %x %x\n", i,len);
				exit(0);
			    }
			}

			pSectionHeaders++;
			
#if 0
			/* not needed, memory is zero */
			if(strcmp(pSectionHeaders[i].Name, ".bss") == 0)
			    memset((void *)LoadAddress, 0,
				   pSectionHeaders[i].Misc.VirtualSize ?
				   pSectionHeaders[i].Misc.VirtualSize :
				   pSectionHeaders[i].SizeOfRawData);
#endif
		}

		_lclose(hFile);

		logstr(lf_log,"imports\n");
		/* we are dependent on other modules, go get and load those */
		if(LoadImportsSection(lpszFileName,BaseAddress, pNTHeader) == 0)
			return 0;

		logstr(lf_log,"exports of %s\n",lpszFileName);
		LoadExportsTable(lpNTModule,pNTHeader,lpszFileName);

		logstr(lf_log,"exec entrypoint of %s\n",lpszFileName);
		lpLibrary = ExecEntryPoint( lpNTModule, lpszFileName, lpszBase);
		logstr(lf_log,"exec entrypoint returns %p\n",lpLibrary);
		return lpLibrary;
	}
	return 0;
}

typedef struct
{
        WORD    flag;
        LPSTR   name;
} WORD_FLAG_DESCRIPTIONS;

typedef struct
{
        DWORD   flag;
        LPSTR   name;
} DWORD_FLAG_DESCRIPTIONS;
#define IMAGE_FILE_DLL                  0x2000

/* Bitfield values and names for the IMAGE_FILE_HEADER flags */
WORD_FLAG_DESCRIPTIONS ImageFileHeaderCharacteristics[] =
{
{ IMAGE_FILE_RELOCS_STRIPPED, "RELOCS_STRIPPED" },
{ IMAGE_FILE_EXECUTABLE_IMAGE, "EXECUTABLE_IMAGE" },
{ IMAGE_FILE_LINE_NUMS_STRIPPED, "LINE_NUMS_STRIPPED" },
{ IMAGE_FILE_LOCAL_SYMS_STRIPPED, "LOCAL_SYMS_STRIPPED" },
{ IMAGE_FILE_MINIMAL_OBJECT, "MINIMAL_OBJECT" },
{ IMAGE_FILE_UPDATE_OBJECT, "UPDATE_OBJECT" },
{ IMAGE_FILE_BYTES_REVERSED_LO, "BYTES_REVERSED_LO" },
{ IMAGE_FILE_32BIT_MACHINE, "32BIT_MACHINE" },
{ IMAGE_FILE_PATCH, "PATCH" },
{ IMAGE_FILE_SYSTEM, "SYSTEM" },
{ IMAGE_FILE_DLL, "DLL" },
{ IMAGE_FILE_BYTES_REVERSED_HI, "BYTES_REVERSED_HI" },
{ IMAGE_FILE_DEBUG_STRIPPED,"DEBUGGING INFO STRIPPED" }
};

#define NUMBER_IMAGE_HEADER_FLAGS (sizeof(ImageFileHeaderCharacteristics) / sizeof(WORD_FLAG_DESCRIPTIONS))

void DumpHeader(PIMAGE_FILE_HEADER pImageFileHeader)
{
        UINT headerFieldWidth = 30;
        UINT i;
        char *szMachine;

        logstr(lf_header,"File Header\n");

        switch( pImageFileHeader->Machine )
        {
                case IMAGE_FILE_MACHINE_I386:   szMachine = "i386"; break;
                case IMAGE_FILE_MACHINE_I860:   szMachine = "i860"; break;
                case IMAGE_FILE_MACHINE_R3000:  szMachine = "R3000"; break;
                case IMAGE_FILE_MACHINE_R4000:  szMachine = "R4000"; break;
                case IMAGE_FILE_MACHINE_ALPHA:  szMachine = "alpha"; break;
                default:    szMachine = "unknown"; break;
        }

        logstr(lf_header,"  %-*s%04X (%s)\n", headerFieldWidth, "Machine:",
                                pImageFileHeader->Machine, szMachine);
        logstr(lf_header,"  %-*s%04X\n", headerFieldWidth, "Number of Sections:",
                                pImageFileHeader->NumberOfSections);
        logstr(lf_header2,"  %-*s%08X\n", headerFieldWidth, "TimeDateStamp:",
                                pImageFileHeader->TimeDateStamp);
        logstr(lf_header2,"  %-*s%08X\n", headerFieldWidth, "PointerToSymbolTable:",
                                pImageFileHeader->PointerToSymbolTable);
        logstr(lf_header2,"  %-*s%08X\n", headerFieldWidth, "NumberOfSymbols:",
                                pImageFileHeader->NumberOfSymbols);
        logstr(lf_header2,"  %-*s%04X\n", headerFieldWidth, "SizeOfOptionalHeader:",
                                pImageFileHeader->SizeOfOptionalHeader);

        logstr(lf_header,"  %-*s%04X\n", headerFieldWidth, "Characteristics:",
                                pImageFileHeader->Characteristics);

        for ( i=0; i < NUMBER_IMAGE_HEADER_FLAGS; i++ )
        {
                if ( pImageFileHeader->Characteristics &
                         ImageFileHeaderCharacteristics[i].flag )
                        logstr(lf_header,"    %s\n", ImageFileHeaderCharacteristics[i].name );
        }
	logstr(lf_header,"\n");
}

/* Bitfield values and names for the DllCharacteritics flags */
WORD_FLAG_DESCRIPTIONS DllCharacteristics[] =
{
{ IMAGE_LIBRARY_PROCESS_INIT, "PROCESS_INIT" },
{ IMAGE_LIBRARY_PROCESS_TERM, "PROCESS_TERM" },
{ IMAGE_LIBRARY_THREAD_INIT, "THREAD_INIT" },
{ IMAGE_LIBRARY_THREAD_TERM, "THREAD_TERM" },
};
#define NUMBER_DLL_CHARACTERISTICS (sizeof(DllCharacteristics) / sizeof(WORD_FLAG_DESCRIPTIONS))

/* Bitfield values and names for the LoaderFlags flags */
DWORD_FLAG_DESCRIPTIONS LoaderFlags[] =
{
{ IMAGE_LOADER_FLAGS_BREAK_ON_LOAD, "BREAK_ON_LOAD" },
{ IMAGE_LOADER_FLAGS_DEBUG_ON_LOAD, "DEBUG_ON_LOAD" }
};
#define NUMBER_LOADER_FLAGS (sizeof(LoaderFlags) / sizeof(DWORD_FLAG_DESCRIPTIONS))

/* Names of the data directory elements that are defined */
char *ImageDirectoryNames[] = {
        "EXPORT", "IMPORT", "RESOURCE", "EXCEPTION", "SECURITY", "BASERELOC",
        "DEBUG", "COPYRIGHT", "GLOBALPTR", "TLS", "LOAD_CONFIG" };

#define NUMBER_IMAGE_DIRECTORY_ENTRYS (sizeof(ImageDirectoryNames)/sizeof(char *))

/*
 * Dump the IMAGE_OPTIONAL_HEADER from a PE file
 */
void DumpOptionalHeader(PIMAGE_OPTIONAL_HEADER optionalHeader)
{
        UINT width = 30;
        char *s;
        UINT i;

        logstr(lf_header,"Optional Header\n");

        logstr(lf_header2,"  %-*s%04X\n", width, "Magic", 
		optionalHeader->Magic);
        logstr(lf_header2,"  %-*s%u.%02u\n", width, "linker version",
                optionalHeader->MajorLinkerVersion,
                optionalHeader->MinorLinkerVersion);

        logstr(lf_header,"  %-*s%X\n", width, "size of code", optionalHeader->SizeOfCode);
        logstr(lf_header,"  %-*s%X\n", width, "size of initialized data",
                optionalHeader->SizeOfInitializedData);
        logstr(lf_header,"  %-*s%X\n", width, "size of uninitialized data",
                optionalHeader->SizeOfUninitializedData);
        logstr(lf_header,"  %-*s%X\n", width, "entrypoint RVA",
                optionalHeader->AddressOfEntryPoint);
        logstr(lf_header,"  %-*s%X\n", width, "base of code", 
		optionalHeader->BaseOfCode);
        logstr(lf_header,"  %-*s%X\n", width, "base of data", 
		optionalHeader->BaseOfData);
        logstr(lf_header,"  %-*s%X\n", width, "image base", 
		optionalHeader->ImageBase);

        logstr(lf_header2,"  %-*s%X\n", width, "section align",
                optionalHeader->SectionAlignment);
        logstr(lf_header2,"  %-*s%X\n", width, "file align", 
		optionalHeader->FileAlignment);
        logstr(lf_header2,"  %-*s%u.%02u\n", width, "required OS version",
                optionalHeader->MajorOperatingSystemVersion,
                optionalHeader->MinorOperatingSystemVersion);
        logstr(lf_header2,"  %-*s%u.%02u\n", width, "image version",
                optionalHeader->MajorImageVersion,
                optionalHeader->MinorImageVersion);
        logstr(lf_header2,"  %-*s%u.%02u\n", width, "subsystem version",
                optionalHeader->MajorSubsystemVersion,
                optionalHeader->MinorSubsystemVersion);
        logstr(lf_header2,"  %-*s%X\n", width, "Reserved1", 
		optionalHeader->Reserved1);

        logstr(lf_header,"  %-*s%X\n", width, "size of image", 
		optionalHeader->SizeOfImage);
        logstr(lf_header,"  %-*s%X\n", width, "size of headers",
                optionalHeader->SizeOfHeaders);

        switch( optionalHeader->Subsystem )
        {
                case IMAGE_SUBSYSTEM_NATIVE: s = "Native"; break;
                case IMAGE_SUBSYSTEM_WINDOWS_GUI: s = "Windows GUI"; break;
                case IMAGE_SUBSYSTEM_WINDOWS_CUI: s = "Windows character"; break;
                case IMAGE_SUBSYSTEM_OS2_CUI: s = "OS/2 character"; break;
                case IMAGE_SUBSYSTEM_POSIX_CUI: s = "Posix character"; break;
                default: s = "unknown";
        }
        logstr(lf_header,"  %-*s%04X (%s)\n", width, "Subsystem",
                        optionalHeader->Subsystem, s);

        logstr(lf_header,"  %-*s%04X\n", width, "DLL flags",
                        optionalHeader->DllCharacteristics);
        for ( i=0; i < NUMBER_DLL_CHARACTERISTICS; i++ )
        {
                if ( optionalHeader->DllCharacteristics &
                         DllCharacteristics[i].flag )
                        logstr(lf_header,"  %s\n", 
				DllCharacteristics[i].name );
        }
        if ( optionalHeader->DllCharacteristics )
                logstr(lf_header,"\n");

        logstr(lf_header2,"  %-*s%X\n", width, "stack reserve size",
                optionalHeader->SizeOfStackReserve);
        logstr(lf_header2,"  %-*s%X\n", width, "stack commit size",
                optionalHeader->SizeOfStackCommit);
        logstr(lf_header2,"  %-*s%X\n", width, "heap reserve size",
                optionalHeader->SizeOfHeapReserve);
        logstr(lf_header2,"  %-*s%X\n", width, "heap commit size",
                optionalHeader->SizeOfHeapCommit);

        logstr(lf_header2,"  %-*s%08X\n", width, "loader flags",
                optionalHeader->LoaderFlags);

        for ( i=0; i < NUMBER_LOADER_FLAGS; i++ )
        {
                if ( optionalHeader->LoaderFlags &
                         LoaderFlags[i].flag )
                        logstr(lf_header,"  %s", LoaderFlags[i].name );
        }

        if ( optionalHeader->LoaderFlags )
                logstr(lf_header,"\n");

        logstr(lf_header,"  %-*s%X\n", width, "RVAs & sizes",
                optionalHeader->NumberOfRvaAndSizes);

        logstr(lf_header,"Data Directory\n");
        for ( i=0; i < optionalHeader->NumberOfRvaAndSizes; i++)
        {
                logstr(lf_header,"  %-12s rva: %08X  size: %08X\n",
                        (i >= NUMBER_IMAGE_DIRECTORY_ENTRYS)
                                ? "unused" : ImageDirectoryNames[i],
                        optionalHeader->DataDirectory[i].VirtualAddress,
                        optionalHeader->DataDirectory[i].Size);
        }
}


/* Bitfield values and names for the IMAGE_SECTION_HEADER flags */
DWORD_FLAG_DESCRIPTIONS SectionCharacteristics[] = 
{
{ IMAGE_SCN_CNT_CODE, "CODE" },
{ IMAGE_SCN_CNT_INITIALIZED_DATA, "INITIALIZED_DATA" },
{ IMAGE_SCN_CNT_UNINITIALIZED_DATA, "UNINITIALIZED_DATA" },
{ IMAGE_SCN_LNK_INFO, "LNK_INFO" },
{ IMAGE_SCN_LNK_OVERLAY, "LNK_OVERLAY" },
{ IMAGE_SCN_LNK_REMOVE, "LNK_REMOVE" },
{ IMAGE_SCN_LNK_COMDAT, "LNK_COMDAT" },
{ IMAGE_SCN_MEM_DISCARDABLE, "MEM_DISCARDABLE" },
{ IMAGE_SCN_MEM_NOT_CACHED, "MEM_NOT_CACHED" },
{ IMAGE_SCN_MEM_NOT_PAGED, "MEM_NOT_PAGED" },
{ IMAGE_SCN_MEM_SHARED, "MEM_SHARED" },
{ IMAGE_SCN_MEM_EXECUTE, "MEM_EXECUTE" },
{ IMAGE_SCN_MEM_READ, "MEM_READ" },
{ IMAGE_SCN_MEM_WRITE, "MEM_WRITE" },
};

#define NUMBER_SECTION_CHARACTERISTICS  (sizeof(SectionCharacteristics) / sizeof(DWORD_FLAG_DESCRIPTIONS))

/*
** Dump the section table from a PE file or an OBJ
*/

void 
DumpSectionTable(void *LoadAddress,PIMAGE_SECTION_HEADER section, unsigned cSections)
{
	unsigned j;

	logstr(lf_header,"Section Table %x %d\n",LoadAddress,cSections);
	
	logstr(lf_header,"  %02X %-8.8s  %s: %08X  VirtAddr:  %08X\n",
			cSections, section->Name, "VirtSize",
			section->Misc.VirtualSize, section->VirtualAddress);
	logstr(lf_header,"    raw data offs:   %08X  raw data size: %08X\n",
			section->PointerToRawData, section->SizeOfRawData );
	logstr(lf_header,"    relocation offs: %08X  relocations:   %08X\n",
			section->PointerToRelocations, section->NumberOfRelocations);
	logstr(lf_header,"    line # offs:     %08X  line #'s:      %08X\n",
			section->PointerToLinenumbers, section->NumberOfLinenumbers );
	logstr(lf_header,"    characteristics: %08X\n", section->Characteristics);

	for ( j=0; j < NUMBER_SECTION_CHARACTERISTICS; j++ )
	{
		if ( section->Characteristics & 
			SectionCharacteristics[j].flag )
			logstr(lf_header,"      %s\n", 
				SectionCharacteristics[j].name );
	}
}

extern void debuggerbreak();
int
Win32Trap(long bogus)
{
	long *backtrack;
	
	/* this is the return address of the caller */
	backtrack = &bogus;
	backtrack--;

	/* now backup to the address in the previous instruction */
	backtrack = (long *) *backtrack;
	backtrack--;

	logstr(LF_CONSOLE,"Win32 Code Segment Trap %p\n",*backtrack);
	debuggerbreak();
	return 0;
}

DWORD
GetProcAddress32(HINSTANCE hInstance,LPSTR lpstr)
{
//TODO
	logstr(LF_CONSOLE,"GetProcAddress(%x,%s)\n",
		hInstance,lpstr);
	return 0;
}

static DWORD
ReturnAddress(int id,char *name,DWORD faddr)
{ 
	if(faddr == 0)
		logstr(lf_log,"function address null case %d %s %d\n",
			id,name,faddr);
	return faddr;
}

DWORD 
PEGetProcAddressByName( MODULE_HEADERS *mp, PIMAGE_IMPORT_BY_NAME pOrdinalName)
{
	PIMAGE_EXPORT_DIRECTORY exportDir;
	DWORD i;
	PDWORD functions;
	PWORD ordinals;
	PSTR *name;
	LPSTR fname;
        IMAGE_DATA_DIRECTORY    dir;
	DWORD	faddress = 0;
	HINSTANCE hInst;

	if (mp->pNTHeader == 0) {

	   if(mp->modulealias && mp->maptable) {
		char *name;
	        hInst = 0;

		if (mp->modulecount == 0) {
			int n = 0;
			for(n=0;mp->maptable[n].lpstr;n++);
			mp->modulecount = n;	
		}

		if(pOrdinalName->Hint > mp->modulecount) {
			
	           logstr(lf_imports,"function exceeds library ordinal %d %s\n",
			pOrdinalName->Hint,pOrdinalName->Name);
			return ReturnAddress(0,pOrdinalName->Name,0);
		}

		name = mp->maptable[pOrdinalName->Hint].lpstr;
		if (name && (strcasecmp(pOrdinalName->Name,name) == 0)) {
			faddress = (DWORD) mp->maptable[pOrdinalName->Hint].faddr;
			return ReturnAddress(1,pOrdinalName->Name,faddress);
		}
		if(name && strlen(name))
	           logstr(lf_imports,"function import mismatch %d %s %s\n",
			 pOrdinalName->Hint,pOrdinalName->Name,name);

		for(i=0;mp->maptable[i].lpstr;i++) {
			name = 	mp->maptable[i].lpstr;
			if (strlen(name) == 0)
			   continue;

			if (strcasecmp(pOrdinalName->Name,name) == 0) {
			    faddress = (DWORD) mp->maptable[i].faddr;

	        	    logstr(lf_log,"relocated %d %s to %d\n",
			 	pOrdinalName->Hint,pOrdinalName->Name,i);
			    return ReturnAddress(2,pOrdinalName->Name,faddress);
			}	
		}
	   }
	   return ReturnAddress(3,pOrdinalName->Name,0);
	}

	/* get the image_data_directory for exported symbols */
	dir = mp->pNTHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT];

	/* convert to pointer */
	exportDir = MAKEPTR((PIMAGE_EXPORT_DIRECTORY),
		mp->BaseAddress,
		dir.VirtualAddress);

	/* any functions at all? */
	if (exportDir->NumberOfFunctions == 0) {
	   return ReturnAddress(4,pOrdinalName->Name,0);
	}

	/* create array of function addresses */
	functions = MAKEPTR((PDWORD),mp->BaseAddress,
		exportDir->AddressOfFunctions);

	/* create array of function ordinals */
	ordinals =  MAKEPTR((PWORD),mp->BaseAddress,
		exportDir->AddressOfNameOrdinals);

	/* create array of RVA's to function names */
	name =      MAKEPTR((PSTR *),mp->BaseAddress,
		exportDir->AddressOfNames);
	
	/* before we do linear search through name table */
	/* use quick lookup first, using the hint...  */
	/* if we have valid hint, get pointer to underlying name */
	if (pOrdinalName->Hint < exportDir->NumberOfNames) {

		fname = MAKEPTR((LPSTR),mp->BaseAddress,name[pOrdinalName->Hint]);
		/* have we matched the strings? */
		if (strcmp((char *)fname,pOrdinalName->Name) == 0) {
			faddress =  MAKEPTR((DWORD),mp->BaseAddress,
				functions[pOrdinalName->Hint]);
	   	        return ReturnAddress(5,pOrdinalName->Name,faddress);
		}
			
		/* no, so do linear search, bummer its wrong??? */
		logstr(lf_header,"requested f=%s ord=%x found %s ord=%x\n",
			(char *)pOrdinalName->Name,
			(int)pOrdinalName->Hint,
			MAKEPTR((LPSTR),mp->BaseAddress,name[pOrdinalName->Hint]),
			ordinals[pOrdinalName->Hint]);
	} else {
		logstr(lf_header,"requested bad entry\n");
	}

	/* resort to straight line linear search */
	/* should/could we ignore comparing strings, until we */
	/* are in range? */
	logstr(lf_log,"searching for function %s\n",pOrdinalName->Name);

	for ( i=0; i < exportDir->NumberOfNames; i++ )
	{
		/* get the pointer to the name */
		fname = MAKEPTR((LPSTR),mp->BaseAddress,*name);

		/* does this one match what we want... */
		if (strcmp((char *)fname,pOrdinalName->Name) == 0)
			faddress = MAKEPTR((DWORD),mp->BaseAddress,*functions);

		/* done with search, break out... */
		/* if(faddress) */
		/*   break; */

		/* bump all the pointers */
		functions++;
		ordinals++;
		name++;		
	}

	/* the name is not in the module... */
	logstr(lf_log,"no functions\n");
	return ReturnAddress(6,pOrdinalName->Name,faddress);
}

/* */
/* Dump the imports table (the .idata section) of a PE file */
/* */
int LoadImportsSection(char *BaseName,void * BaseAddress, PIMAGE_NT_HEADERS pNTHeader)
{
	PIMAGE_IMPORT_DESCRIPTOR importDesc;
	PIMAGE_THUNK_DATA thunk;
	PIMAGE_IMPORT_BY_NAME pOrdinalName;
        IMAGE_DATA_DIRECTORY    dir;
	char *libname;
	MODULE_HEADERS *mp;
	DWORD faddr;
	DWORD *ip;
	char *ModuleName = BaseName;
		
	dir=pNTHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT];
	importDesc = MAKEPTR((PIMAGE_IMPORT_DESCRIPTOR),
		BaseAddress,
		dir.VirtualAddress);

 	logstr(lf_imports,"  DLL Imports Table: %s\n", ModuleName);

	while ( 1 )
	{
		/* See if we've reached an empty IMAGE_IMPORT_DESCRIPTOR */
		if ( (importDesc->TimeDateStamp==0 ) && (importDesc->Name==0) )
			break;

		libname = (PBYTE)(importDesc->Name) + (int)BaseAddress;

		logstr(lf_header,"  Library: %s\n",libname); 
		mp = ExecPE(libname, BaseName);

	        if (mp == 0) {
		   logstr(lf_log,"ExecPE failed: %s\n",libname); 
		   return 0;
		}

		logstr(lf_header,"  Library: %s\n",libname); 

		logstr(lf_header2,"  Hint/Name Table: %08X\n", 
			importDesc->u.Characteristics);
 		logstr(lf_header2,"  TimeDateStamp:   %08X\n", 
			importDesc->TimeDateStamp);
 		logstr(lf_header2,"  ForwarderChain:  %08X\n", 
			importDesc->ForwarderChain);
 		logstr(lf_header2,"  First thunk RVA: %08X\n", 
			importDesc->FirstThunk);
	
		thunk = MAKEPTR((PIMAGE_THUNK_DATA),BaseAddress,importDesc->u.Characteristics);
		ip    = MAKEPTR((DWORD *),BaseAddress,importDesc->FirstThunk);

		while (1) {
			if (thunk->u1.AddressOfData == 0)
			   break;

			pOrdinalName = thunk->u1.AddressOfData;

			thunk++;	

			if((int) pOrdinalName & 0x80000000) {
			     char ordstr[80];
			     sprintf(ordstr,"ordinal 0x%x of %s",
			          (int) pOrdinalName & 0x7fffffff,libname);

		             logstr(lf_imports,"ordinal import %4.4p %s\t@%p\n",
					pOrdinalName,ordstr,ip);
			     faddr = (DWORD) setstub(ordstr,
						(int)pOrdinalName);
			     *ip = faddr;
			} else {

			     pOrdinalName = (PIMAGE_IMPORT_BY_NAME) 
				     ((PBYTE) pOrdinalName + (int)BaseAddress);

			     faddr = PEGetProcAddressByName(mp,pOrdinalName);

			     logstr(lf_log,
				"    import %-24s %s\t@%p %4d %s\n",
				pOrdinalName->Name,libname,ip, 
				pOrdinalName->Hint,faddr?"":"TRAP");

			      if (faddr == 0) {
				faddr = (DWORD) setstub(pOrdinalName->Name,
						pOrdinalName->Hint);
			      }
			     *ip = faddr;
			}
			ip++;
		}

		importDesc++;	/* advance to next IMAGE_IMPORT_DESCRIPTOR */
	}
	return 1;
}

void 
LoadExportsTable(MODULE_HEADERS *mp,PIMAGE_NT_HEADERS pNTHeader,char *ModuleName)
{
	PIMAGE_EXPORT_DIRECTORY exportDir;
	DWORD i;
	PDWORD functions;
	PWORD ordinals;
	PSTR *name;
	LPSTR fname;
        IMAGE_DATA_DIRECTORY    dir;


	/* get the image_data_directory for exported symbols */
	dir = pNTHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT];

	/* convert to pointer */
	exportDir = MAKEPTR((PIMAGE_EXPORT_DIRECTORY),
		mp->BaseAddress,
		dir.VirtualAddress);

	/* any functions at all? */
	if (exportDir->NumberOfFunctions == 0)
	   return ;

	logstr(lf_exports,"Exports of %s\n",ModuleName);

	/* create array of function addresses */
	functions = MAKEPTR((PDWORD),mp->BaseAddress,
		exportDir->AddressOfFunctions);

	/* create array of function ordinals */
	ordinals =  MAKEPTR((PWORD),mp->BaseAddress,
		exportDir->AddressOfNameOrdinals);

	/* create array of RVA's to function names */
	name =      MAKEPTR((PSTR *),mp->BaseAddress,
		exportDir->AddressOfNames);
	
	/* resort to straight line linear search */
	/* should/could we ignore comparing strings, until we */
	/* are in range? */
	for ( i=0; i < exportDir->NumberOfNames; i++ )
	{
		/* get the pointer to the name */
		fname = MAKEPTR((LPSTR),mp->BaseAddress,*name);

	        logstr(lf_exports,"export %x %s\n",
			*ordinals,fname);

		/* bump all the pointers */
		functions++;	/* function address */
		ordinals++;	/* function ordinal */
		name++;		/* function name */
	}

	/* the name is not in the module... */
	return ;
}

extern WORD AssignSelector(LPBYTE , WORD , BYTE , DWORD );

MODULE_HEADERS *
ExecEntryPoint(MODULE_HEADERS *lpNTModule ,char *lpszModuleName, char *lpszBase)
{
	void *BaseAddress; 
	PIMAGE_NT_HEADERS pNTHeader;
	int *fp;
	/*static*/ THDB *initthread;
	WORD	    wSel;
	FARPROC	    entry;
	HGLOBAL       hMem;

	 if (lpszBase == 0) {
		   HINSTANCE	hInst;
		   HMODULE      hModule;
		   MODULEINFO   *modinfo;
		   extern TWINRT DefaultTwinRT32;

		   hInst = TWIN_GetInstanceFromTask( GetCurrentTask());
		   hModule = GetModuleFromInstance(hInst);
		   modinfo = GETMODULEINFO(hModule);

		   if(modinfo) {
			modinfo->dwWin32flag = 1;
			modinfo->lpNTHeader  = (void *) lpNTModule;
    			modinfo->lptwinrt    = &DefaultTwinRT32;
		   }
	}

	CurrentProcess = lpNTModule;
	BaseAddress = lpNTModule->BaseAddress;
	pNTHeader   = lpNTModule->pNTHeader;

	fp = MAKEPTR((int *), BaseAddress, 
			pNTHeader->OptionalHeader.AddressOfEntryPoint);

	logstr(lf_log,"execute: mod=%s address(%x) farproc=%x\n",
		lpszModuleName,
		pNTHeader->OptionalHeader.AddressOfEntryPoint,
		fp);

	if(fp == 0)
		return lpNTModule;

	/* allocate a block for this one, and execute it */
	hMem = GlobalAlloc(0,sizeof(THDB));
	initthread = (THDB *) GlobalLock(hMem);

	wSel = AssignSelector((LPBYTE) initthread, 0, TRANSFER_DATA32, sizeof(THDB));

	logstr(LF_CONSOLE,"invoke_binary32(%s,%p) address=%x\n",
		lpszModuleName,fp,
		pNTHeader->OptionalHeader.AddressOfEntryPoint);

	if(setjmp(lpNTModule->env) == 0) {

		__asm__ __volatile__("movw %w0,%%fs\t\n"::"r" (wSel));
		entry = (FARPROC) fp;
		entry();

	}

	logstr(LF_CONSOLE,"invoke_binary32: returns %d\n",
		lpNTModule->exit_code);
	logstr(lf_log,"execute: %s returns %d\n",
		lpszModuleName,lpNTModule->exit_code);
		

	return lpNTModule;
}

void
ExitProcess(int id)
{
	logstr(lf_log,"ExitProcess(%d)\n",id);
	CurrentProcess->exit_code = id;
	longjmp(CurrentProcess->env,0);
}

long
LibMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR lpCmdLine, int nCmdShow)
{
	logstr(lf_log,"LibMain(HINSTANCE %x, LPSTR %p, int %d)\n",
		hInst,hPrevInst,lpCmdLine,nCmdShow);

	return 0;
}
