
#include <stdio.h>

#include "windows.h"
#include "Log.h"
#include "winnt.h"
#include "heap.h"
#include "RscData.h"
#include "stdarg.h"
#include "Resources.h"
#include "Module.h"
#include "lib32.h"
#include "WinConfig.h"
#include "Strings32.h"

char * GetTwinMsgCode(HWND ,int );

// based on current running application
int
GetStartupInfoA(LPSTARTUPINFO32A lp)
{
	APISTR((LF_API,"GetStartupInfoA(%p) sizeof=%x\n",
		lp,sizeof(STARTUPINFO32A)));
	memset(lp,0,sizeof(STARTUPINFO32A));
	lp->dwFlags = 1;
	lp->wShowWindow = 1;	
	return 0;
}

DWORD
GetFileType32(HANDLE hFile)
{
	APISTR((LF_API,"GetFileType32(%x)\n",hFile));
	return 0;	
}

HANDLE
GetStdHandle(DWORD nStdHandle)
{
	APISTR((LF_API,"GetStdHandle(%x)\n",nStdHandle));
	if(nStdHandle == STD_INPUT_HANDLE)
		return INVALID_HANDLE_VALUE;
	if(nStdHandle == STD_OUTPUT_HANDLE)
		return INVALID_HANDLE_VALUE;
	if(nStdHandle == STD_ERROR_HANDLE)
		return INVALID_HANDLE_VALUE;

	return INVALID_HANDLE_VALUE;
}

UINT
GetACP(void)
{
	/* return ansi code page 1252 */
	return 0x4e4;  
}

typedef struct {
	UINT   MaxCharSize;
	BYTE   DefaultChar[2];
	BYTE   LeadByte[8];
} CPINFO;

typedef CPINFO *LPCPINFO;

UINT
GetCPInfo(UINT codepage,LPCPINFO lpcpinfo)
{
	APISTR((LF_API,"GetCPInfo(%x,%p)\n",codepage,lpcpinfo));

	lpcpinfo->DefaultChar[0] = '?';
	switch (codepage)
	{
	case 932 : /* Shift JIS (japan) */
		lpcpinfo->MaxCharSize = 2;
		lpcpinfo->LeadByte[0]= 0x81; lpcpinfo->LeadByte[1] = 0x9F;
		lpcpinfo->LeadByte[2]= 0xE0; lpcpinfo->LeadByte[3] = 0xFC;
		lpcpinfo->LeadByte[4]= 0x00; lpcpinfo->LeadByte[5] = 0x00;
		break;
	case 936 : /* GB2312 (Chinese) */
	case 949 : /* KSC5601-1987 (Korean) */
	case 950 : /* BIG5 (Chinese) */
		lpcpinfo->MaxCharSize = 2;
		lpcpinfo->LeadByte[0]= 0x81; lpcpinfo->LeadByte[1] = 0xFE;
		lpcpinfo->LeadByte[2]= 0x00; lpcpinfo->LeadByte[3] = 0x00;
		break;
	case 1361 : /* Johab (Korean) */
		lpcpinfo->MaxCharSize = 2;
		lpcpinfo->LeadByte[0]= 0x84; lpcpinfo->LeadByte[1] = 0xD3;
		lpcpinfo->LeadByte[2]= 0xD8; lpcpinfo->LeadByte[3] = 0xDE;
		lpcpinfo->LeadByte[4]= 0xE0; lpcpinfo->LeadByte[5] = 0xF9;
		lpcpinfo->LeadByte[6]= 0x00; lpcpinfo->LeadByte[7] = 0x00;
	break;
	default :
		lpcpinfo->MaxCharSize = 1;
		lpcpinfo->LeadByte[0]= 0x00; lpcpinfo->LeadByte[1] = 0x00;
	break;
	}

	return codepage;  
}

LPTSTR
GetEnvironmentStrings(VOID)
{
	LPTSTR penv;
	APISTR((LF_APISTUB,"GetEnvironmentsStringsW\n"));
	penv = (LPTSTR)  malloc(64*1024);
	//BuildEnvBuffer(GetCommandLineA(),penv);
	penv[0] = 0;
	penv[1] = 0;
	return penv;
}

LPTSTR
GetEnvironmentStringsW(VOID)
{
	return GetEnvironmentStrings();
}
int
FreeEnvironmentStringsW(LPSTR penv)
{
	APISTR((LF_API,"FreeEnvironmentsStringsW(%p)\n",penv));
	free(penv);
	return 1;
}

int
GetModuleFileNameA(HINSTANCE hInstance,LPSTR lpstr,int len)
{
	char string[80];

	if(HIWORD(lpstr) == 0) {
		wsprintf(string,"0x%x",lpstr);
		lpstr = string;
	}
	APISTR((LF_API,"GetModuleFileNameA(%d,%s,%d)\n",
		hInstance,lpstr,len));

	return (int) strncpy(lpstr,"win32.exe",min(10,len));
}

HTASK GetCurrentTask();
HINSTANCE TWIN_GetInstanceFromTask( HTASK );

HINSTANCE
GetModuleHandleA(LPCTSTR lpModuleName)
{
	HINSTANCE	hInst;

	APISTR((LF_API,"GetModuleHandleA(%s)\n",
		lpModuleName?lpModuleName:""));

	hInst = TWIN_GetInstanceFromTask( GetCurrentTask());

	return hInst;
}

HMODULE GetModuleFromInstance(HINSTANCE);	

LPTSTR
GetCommandLineA(VOID)
{
	LPMODULEINFO modinfo;
	HMODULE     hModule;
	HINSTANCE	hInst;
	char *lpstr;

	APISTR((LF_API,"GetCommandLineA\n"));

	hInst = TWIN_GetInstanceFromTask( GetCurrentTask());
	hModule = GetModuleFromInstance(hInst);	
	modinfo = GETMODULEINFO(hModule);
	
	lpstr = strchr(modinfo->lpFileName,' ');
	if(lpstr) lpstr++;
	else      lpstr = modinfo->lpFileName;

	return lpstr;
}

int
IsBadReadPtr32(void *p,int nsize)
{
	APISTR((LF_API,"IsBadReadPtr32(%p,%d)\n",p,nsize));
	return 0;
}

int
IsBadWritePtr32(void *p,int nsize)
{
	APISTR((LF_API,"IsBadWritePtr32(%p,%d)\n",p,nsize));
	return 0;
}

int
GetVersion32()
{
	APISTR((LF_API,"GetVersion32 %x\n",0x05650004));
	return 0x05650004;
}


HWND
CreateWindowExA(DWORD dwExStyle, LPCSTR lpClass, 
		LPCSTR lpWindowName, DWORD dwStyle,
		int X, int Y, int nWidth, int nHeight, 
		HWND hWndParent, HMENU hMenu, HINSTANCE hInstance, 
		LPVOID lpParam)
{
	APISTR((LF_API,"CreateWindowExA(...)\n"));
	return CreateWindowEx(dwExStyle,lpClass,lpWindowName,dwStyle,
		X,Y,nWidth,nHeight,hWndParent,hMenu,hInstance,lpParam);
}

HWND
CreateWindowExW(DWORD dwExStyle, LPCSTR lpClass, 
		LPCSTR lpWindowName, DWORD dwStyle,
		int X, int Y, int nWidth, int nHeight, 
		HWND hWndParent, HMENU hMenu, HINSTANCE hInstance, 
		LPVOID lpParam)
{
	char lpclassname[80];
	char lpwinname[80];

	lpclassname[0] = 0;
	lpwinname[0] = 0;
	WideCharToMultiByte(0,0,(LPCWSTR)lpClass,0,lpclassname,80,0,0);
	WideCharToMultiByte(0,0,(LPCWSTR)lpWindowName,0,lpwinname,80,0,0);


	APISTR((LF_API,"CreateWindowExW(...)\n"));
	return CreateWindowEx(dwExStyle,lpclassname,lpwinname,dwStyle,
		X,Y,nWidth,nHeight,hWndParent,hMenu,hInstance,lpParam);
}

BOOL    
GetMessageA(MSG *lpmsg, HWND hWnd, UINT idlo, UINT idhi)
{
	APISTR((LF_API,"GetMessageA(%p,%x,%x,%x)\n",
		lpmsg,hWnd,idlo,idhi));

	return GetMessage(lpmsg,hWnd,idlo,idhi);
}

BOOL    
GetMessageW(MSG *lpmsg, HWND hWnd, UINT idlo, UINT idhi)
{
	APISTR((LF_API,"GetMessageW(%p,%x,%x,%x)\n",
		lpmsg,hWnd,idlo,idhi));

	return GetMessage(lpmsg,hWnd,idlo,idhi);
}

int
TranslateAcceleratorA(HWND hWnd, HACCEL hAccel, MSG *lpmsg)
{
	APISTR((LF_API,"TranslateAcceleratorA(%x,%x,%p)\n",
		hWnd,hAccel,lpmsg));

	return TranslateAccelerator(hWnd,hAccel,lpmsg);
}

int
TranslateAcceleratorW(HWND hWnd, HACCEL hAccel, MSG *lpmsg)
{
	APISTR((LF_API,"TranslateAcceleratorW(%x,%x,%p)\n",
		hWnd,hAccel,lpmsg));

	return TranslateAccelerator(hWnd,hAccel,lpmsg);
}

LONG    
DispatchMessageA(MSG *lpmsg)
{
	APISTR((LF_API,"DispatchMessageA(%p)\n",lpmsg));
	return DispatchMessage(lpmsg);

}

LONG    
DispatchMessageW(MSG *lpmsg)
{
	APISTR((LF_API,"DispatchMessageW(%p)\n",lpmsg));
	return DispatchMessage(lpmsg);

}

DWORD
DefWindowProcA(HWND hWnd,UINT msg,DWORD wparam,DWORD lparam)
{
	APISTR((LF_API,"DefWindowProcA(%x,%x:%s,%x,%x)\n",
		hWnd,msg,
		GetTwinMsgCode(hWnd, msg),
		wparam,lparam));
	return DefWindowProc(hWnd,msg,wparam,lparam);
}

DWORD
DefWindowProcW(HWND hWnd,UINT msg,DWORD wparam,DWORD lparam)
{
	APISTR((LF_API,"DefWindowProcW(%x,%x:%s,%x,%x)\n",
		hWnd,msg,
		GetTwinMsgCode(hWnd, msg),
		wparam,lparam));
	return DefWindowProc(hWnd,msg,wparam,lparam);
}

DWORD
DefDlgProcA(HWND hDlg,UINT msg,DWORD wparam,DWORD lparam)
{
	APISTR((LF_API,"DefDlgProcA(%x,%x:%s,%x,%x)\n",
		hDlg,msg,
		GetTwinMsgCode(hDlg, msg),
		wparam,lparam));
	return DefDlgProc(hDlg,msg,wparam,lparam);

}

DWORD
DefDlgProcW(HWND hWnd,UINT msg,DWORD wparam,DWORD lparam)
{
	APISTR((LF_API,"DefDlgProcW(%x,%x:%s,%x,%x)\n",
		hWnd,msg,
		GetTwinMsgCode(hWnd, msg),
		wparam,lparam));
	return DefDlgProc(hWnd,msg,wparam,lparam);
}

static void
FixClassEx(WNDCLASSEX *wc,char *MenuName,char *ClassName)
{
	
	WideCharToMultiByte(0,0,(LPCWSTR)wc->lpszMenuName,0,MenuName,32,0,0);
	wc->lpszMenuName = MenuName;

	WideCharToMultiByte(0,0,(LPCWSTR)wc->lpszClassName,0,ClassName,32,0,0);
	wc->lpszClassName = ClassName;
#if 0
	APISTR((LF_DEBUG,"WNDCLASSEX(%p)\n",wc));
	APISTR((LF_DEBUG,"sz=%d Style=%x xtra=%d,%d\n",
		wc->cbSize,wc->style,wc->cbClsExtra,
		wc->cbWndExtra));
	APISTR((LF_DEBUG,"inst=%x icon=%x cursor=%x\n",
		wc->hInstance,wc->hIcon,wc->hCursor));
	APISTR((LF_DEBUG,"bg=%x iconsm=%x\n",
		wc->hbrBackground,wc->hIconSm));
	DumpMemory("MenuName",wc->lpszMenuName,8);
	DumpMemory("ClassName",wc->lpszClassName,8);
#endif
}

int
RegisterClassA(WNDCLASS *wc)
{
	APISTR((LF_API,"RegisterClassA(%p)\n",wc));
	return RegisterClass(wc);	
}

int
RegisterClassW(WNDCLASS *wc)
{
	char MenuName[32];
	char ClassName[32];

	APISTR((LF_API,"RegisterClassW(%p)\n",wc));
	WideCharToMultiByte(0,0,(LPCWSTR)wc->lpszMenuName,0,MenuName,32,0,0);
	wc->lpszMenuName = MenuName;

	WideCharToMultiByte(0,0,(LPCWSTR)wc->lpszClassName,0,ClassName,32,0,0);
	wc->lpszClassName = ClassName;

	APISTR((LF_API,"RegisterClassW(%s)(%s)\n",
		MenuName, ClassName));
	return RegisterClass(wc);	
}
int
RegisterClassExA(WNDCLASSEX *wc)
{
	APISTR((LF_API,"RegisterClassExA(%p)\n",wc));
	return RegisterClassEx(wc);	
}

int
RegisterClassExW(WNDCLASSEX *wc)
{
	char MenuName[32];
	char ClassName[32];
	APISTR((LF_API,"RegisterClassExW(%p)\n",wc));
	FixClassEx(wc,MenuName,ClassName);
	return RegisterClassEx(wc);	
}

void
ShellAboutA(HWND hWnd, LPCSTR lpszCaption, LPCSTR lpszAboutText, HICON hIcon)
{
	FARPROC		shellabout;
	HINSTANCE	hInst;
	LPCSTR lpstr = lpszCaption;
	LPCSTR lptxt = lpszAboutText;

	APISTR((LF_API,"ShellAboutA(HWND=%x,Caption=%p,Text=%p,HICON=%x)\n",
		hWnd,lpszCaption,lpszAboutText,hIcon));

	/* LATER: Move our icon to RscData.rc instead */
	hIcon = LoadIcon ( 0, MAKEINTRESOURCE ( IDI_WILLOWS ) );

	/* note: this will fail if we are using the REAL shell.dll */
	/*       so if we don't have shellabout, then call         */
	/*       MessageBox instead.                               */
	/* note: we need to do this, rather than a direct call to  */	
	/* 	 shell about function, it is now in a library      */
	/*       It does not have a native to binary hook 	   */
	hInst = LoadLibrary("shell");
	shellabout = GetProcAddress(hInst,"ShellAbout");

	/*								*/
 	/* Note: this will fail even if the real shell.dll is loaded 	*/
	/*	 because we don't have an OEM table for the shell       */
	/*	 functions.  A binary application can call the function */
	/*	 directly.						*/

	if(lpstr == 0)
		lpstr = "Willows TWIN/NT";
	if(lptxt == 0)
		lptxt = "Willows Software, Inc.";

	if (shellabout)
		shellabout( hWnd, lpstr,lptxt, hIcon);
	else 
		MessageBox(hWnd, lptxt, lpstr, MB_OK);
	
	if(hIcon)
		DestroyIcon ( hIcon );

}

void
ShellAboutW(HWND hWnd, LPWSTR lpszCaption, LPWSTR lpszAboutText, HICON hIcon)
{
	char *lpcap;
	char *lptxt;
	int   sz;
	sz = lstrlenW(lpszCaption);
	lpcap = alloca(sz);
	sz = lstrlenW(lpszAboutText);
	lptxt = alloca(sz);
	lstrcpyWtoA(lpcap,lpszCaption);
	lstrcpyWtoA(lptxt,lpszAboutText);

	ShellAboutA(hWnd, lpcap, lptxt, hIcon);
}


void
OutputDebugStringA(char *lpstr)
{
	APISTR((LF_CONSOLE,lpstr));
}

static LCID	UserLCID;
static LCID	SystemLCID;

LCID
GetUserDefaultLCID()
{
    	if ( !UserLCID )
		UserLCID = GetPrivateProfileInt("boot", "LCID",
					0x409, GetTwinFilename());
	APISTR((LF_API,"GetUserDefaultLCID()\n",UserLCID));
	return UserLCID;
}
LCID
GetSystemDefaultLCID()
{
    	if ( !SystemLCID )
		SystemLCID = GetPrivateProfileInt("boot", "LCID",
					0x409, GetTwinFilename());
	APISTR((LF_API,"GetSystemDefaultLCID()\n",SystemLCID));
	return SystemLCID;
}

//BOOL
//RegisterHotKey32(HWND hWnd, int id, UINT modifier,UINT vkkey)
//{
	//APISTR((LF_API,"RegisterHotKey32(%x,%x,%x,%x,%x)\n",
		//hWnd,id,modifier,vkkey));
	//return RegisterHotKey(hWnd,id,modifier,vkkey);
//}
DWORD
LocalAlloc32(UINT flags,UINT bytes)
{
	HANDLE hMem;
	DWORD  dwWord;
	APISTR((LF_APICALL,"LocalAlloc(%x,%x)\n", flags,bytes));
	
	hMem = GlobalAlloc(flags,bytes);
	if(hMem && !(flags  & LMEM_MOVEABLE)) {
		dwWord = (DWORD) GlobalLock(hMem);
		APISTR((LF_APIRET,"LocalAlloc returns DWORD (ptr) %x\n", 
			dwWord));
		return dwWord;
	}

	APISTR((LF_APIRET,"LocalAlloc returns DWORD (handle) %x\n", hMem));
	return hMem;
}

DWORD
LocalReAlloc32(DWORD hLocal,UINT flags,UINT bytes)
{
	HANDLE hMem = hLocal;
	DWORD  dwWord;
	
	APISTR((LF_APICALL,"LocalReAlloc(%x,%x,%x)\n", hLocal,flags,bytes));
	if(HIWORD(hLocal))
		hMem = GlobalHandle32((LPCVOID)hLocal);
		
	hMem =  GlobalReAlloc(hMem,flags,bytes);
	if(HIWORD(hLocal)) {
		dwWord = (DWORD) GlobalLock(hMem);
		APISTR((LF_APIRET,"LocalReAlloc returns DWORD (ptr) %x\n", 
			dwWord));
		return dwWord;
	}	
	APISTR((LF_APIRET,"LocalReAlloc returns DWORD (handle) %x\n", hMem));
	return hMem;
}

HLOCAL
LocalFree32(HLOCAL hLocal)
{
	HANDLE hMem = hLocal;
	HGLOBAL hFree;
	APISTR((LF_APICALL,"LocalFree32(%x)\n", hMem));
	if(HIWORD(hLocal))
		hMem = GlobalHandle32((LPCVOID)hLocal);
		
	hFree =  GlobalFree(hMem);
	APISTR((LF_APIRET,"LocalFree32 returns HLOCAL\n", hFree));
	return (HLOCAL) hFree;
}

LPVOID
LocalLock32(HLOCAL hLocal)
{
	HANDLE hMem = hLocal;
	LPVOID  lpVoid;
	
	APISTR((LF_APICALL,"LocalLock(%x)\n", hLocal));
	lpVoid = (LPVOID) GlobalLock(hMem);
	APISTR((LF_APIRET,"LocalLock returns DWORD (ptr) %p\n", lpVoid));
	return lpVoid;
}

BOOL
GetTextMetricsW(HDC hDC, LPTEXTMETRIC lpMetrics)
{
    BOOL rc;
    APISTR((LF_APICALL,"GetTextMetricsW(HDC=%x,LPTEXTMETRIC=%p)\n",hDC,lpMetrics));
    rc = GetTextMetrics(hDC,lpMetrics);
    APISTR((LF_APIRET,"GetTextMetricsW returns BOOL %d\n",rc));
    return rc;
}

BOOL WINAPI
GetTextMetricsA(HDC hDC, LPTEXTMETRIC lpMetrics)
{
    BOOL rc;
    APISTR((LF_APICALL,"GetTextMetricsA(HDC=%x,LPTEXTMETRIC=%p)\n",hDC,lpMetrics));
    rc = GetTextMetrics(hDC,lpMetrics);
    APISTR((LF_APIRET,"GetTextMetricsA returns BOOL %d\n",rc));
    return rc;
}

char *
MakeTempString(LPWSTR lpwstr,char *alt)
{
	static char   *lpstr;
	static int    current;
	int    len;
	
	len = lstrlenW(lpwstr);
	if(len == 0)
		return "";

	if(!alt && len > current) {
		if(lpstr)
			WinFree(lpstr);
		lpstr = WinMalloc(len);
		current = len;
	}
	if(alt)
		lpstr = alt;
	if(lpstr)
		WideCharToMultiByte(0,0,(LPCWSTR)lpwstr,0,lpstr,len,0,0);
	
	return lpstr;
}
BOOL GetTextExtentPoint32A(HDC hDC,LPCSTR lpcstr,int cbstring,LPSIZE lpSize)
{
	APISTR((LF_API,"GetTextExtentPoint32A(HDC=%x,%s,UINT=%x)\n",
		hDC,lpcstr,cbstring));
	return GetTextExtentPoint(hDC,(LPSTR)lpcstr,cbstring,lpSize);
}

BOOL GetTextExtentPoint32W(HDC hDC,LPCWSTR lpcstr,int cbstring,LPSIZE lpSize)
{
	APISTR((LF_API,"GetTextExtentPoint32W(HDC=%x,%s,UINT=%x)\n",
		hDC,lpcstr,cbstring));
	return GetTextExtentPoint(hDC,(LPSTR)lpcstr,cbstring,lpSize);
}

BOOL GetTextExtentPointA(HDC hDC,LPCSTR lpcstr,int cbstring,LPSIZE lpSize)
{
	APISTR((LF_API,"GetTextExtentPointA(HDC=%x,%s,UINT=%x)\n",
		hDC,lpcstr,cbstring));
	return GetTextExtentPoint(hDC,lpcstr,cbstring,lpSize);
}

BOOL GetTextExtentPointW(HDC hDC,LPCWSTR lpcstr,int cbstring,LPSIZE lpSize)
{
	char *lpstr;
	lpstr = MakeTempString((LPWSTR)lpcstr,0);
	APISTR((LF_API,"GetTextExtentPointW(HDC=%x,%s,UINT=%x)\n",
		hDC,lpstr,cbstring));
	return GetTextExtentPoint(hDC,lpstr,cbstring,lpSize);
}

void
SetDlgItemTextA(HWND hDlg,int nID,char *p)
{
	APISTR((LF_API,"SetDlgItemTextA(%x,%d,%s)\n",
		hDlg,nID,p));
	//DumpMemory("SetDlgItemTextA",p,16);
	SetDlgItemText(hDlg,nID,p);
}

void
SetDlgItemTextW(HWND hDlg,int nID,LPWSTR p)
{
	LPSTR lpstr;
	lpstr = MakeTempString(p,0);
	APISTR((LF_API,"SetDlgItemTextW(%x,%d,%s)\n",
		hDlg,nID,lpstr));
	//DumpMemory("SetDlgItemTextW",p,16);
	SetDlgItemText(hDlg,nID,lpstr);
}

BOOL
TextOutA(HDC hDC,int x,int y,LPCSTR lpstr,int cbstring)
{
	APISTR((LF_API,"TextOutA(HDC=%x,%d,%d],%s,UINT=%x)\n",
		hDC,x,y,lpstr,cbstring));

	return TextOut(hDC,x,y,lpstr,cbstring);
}

BOOL
TextOutW(HDC hDC,int x,int y,LPCSTR lpstr,int cbstring)
{	
	char *tstr;
	tstr = MakeTempString((LPWSTR)lpstr,0);
	APISTR((LF_API,"TextOutW(HDC=%x,%d,%d,%s[%s],UINT=%x)\n",
		hDC,x,y,lpstr,tstr,cbstring));
	return TextOut(hDC,x,y,tstr,cbstring);
}
BOOL
ExtTextOutA(HDC hDC, int x, int y, UINT uiFlags,
	   const RECT *lpRect, LPCSTR lpStr, UINT nCnt, LPINT lpDX)
{
	APISTR((LF_API,"ExtTextOutA(HDC=%x,%d,%d,%d,%p,%s,UINT=%x,%p)\n",
		hDC,x,y,uiFlags,lpRect,lpStr,nCnt,lpDX));
	return ExtTextOut(hDC,x,y,uiFlags,lpRect,lpStr,nCnt,lpDX);
}

BOOL
ExtTextOutW(HDC hDC, int x, int y, UINT uiFlags,
	   const RECT *lpRect, LPCSTR lpStr, UINT nCnt, LPINT lpDX)
{
	char *tstr;
	tstr = MakeTempString((LPWSTR)lpStr,0);
	APISTR((LF_API,"ExtTextOutW(HDC=%x,%d,%d,%d,%p,%s[%s],UINT=%x,%p)\n",
		hDC,x,y,uiFlags,lpRect,lpStr,tstr,nCnt,lpDX));
	return ExtTextOut(hDC,x,y,uiFlags,lpRect,tstr,nCnt,lpDX);
}

LPARAM
SendMessageA(HWND hWnd,UINT msg,WPARAM wparam,LPARAM lparam)
{
	return SendMessage(hWnd,msg,wparam,lparam);
}

LPARAM
SendMessageW(HWND hWnd,UINT msg,WPARAM wparam,LPARAM lparam)
{
	return SendMessage(hWnd,msg,wparam,lparam);
}

BOOL
IsDialogMessageA(HWND hWnd,LPMSG lpmsg)
{
	return IsDialogMessage(hWnd,lpmsg);
}

BOOL
IsDialogMessageW(HWND hWnd,LPMSG lpmsg)
{
	return IsDialogMessage(hWnd,lpmsg);
}

HFONT 
CreateFontW(
    int nHeight,	// logical height of font 
    int nWidth,	// logical average character width 
    int nEscapement,	// angle of escapement 
    int nOrientation,	// base-line orientation angle 
    int fnWeight,	// font weight 
    DWORD fdwItalic,	// italic attribute flag 
    DWORD fdwUnderline,	// underline attribute flag 
    DWORD fdwStrikeOut,	// strikeout attribute flag 
    DWORD fdwCharSet,	// character set identifier 
    DWORD fdwOutputPrecision,	// output precision 
    DWORD fdwClipPrecision,	// clipping precision 
    DWORD fdwQuality,	// output quality 
    DWORD fdwPitchAndFamily,	// pitch and family 
    LPCTSTR lpszFace 	// pointer to typeface name string 
   )
{
  return CreateFont(
	    nHeight,
	    nWidth,	
	    nEscapement,	
	    nOrientation,
	    fnWeight,
	    fdwItalic,	
	    fdwUnderline,
	    fdwStrikeOut,	
	    fdwCharSet,	
	    fdwOutputPrecision,	
	    fdwClipPrecision,	
	    fdwQuality,	
	    fdwPitchAndFamily,	
	    MakeTempString((LPWSTR)lpszFace,0));
}

BOOL
CreateFontIndirectA(LPLOGFONT lpfont)
{
	return CreateFontIndirect(lpfont);
}
BOOL
CreateFontIndirectW(LPLOGFONT lpfont)
{
	return CreateFontIndirect(lpfont);
}

BOOL
WinHelpA( HWND hWnd, LPCSTR lpszHelpFile, UINT fuCommand, DWORD dwData )
{
	APISTR((LF_API,"WinHelpA(%x,%s,%d,%d)\n",
		hWnd,lpszHelpFile,fuCommand,dwData));
	return WinHelp( hWnd, lpszHelpFile, fuCommand, dwData );
}

BOOL
WinHelpW( HWND hWnd, LPCSTR lpszHelpFile, UINT fuCommand, DWORD dwData )
{
	char *lpstr;
	lpstr = MakeTempString((LPWSTR)lpszHelpFile,0);
	APISTR((LF_API,"WinHelpW(%x,%s:%s,%d,%d)\n",
		hWnd,lpszHelpFile,lpstr,fuCommand,dwData));
	return WinHelp( hWnd, lpstr, fuCommand, dwData );
}

BOOL GetTextExtentExPointA(HDC hDC,LPCTSTR lpsz,int cch,int nMax,LPINT lpfit,LPINT lpdx, LPSIZE lpsize)
{
	return GetTextExtentExPoint(hDC,lpsz,cch,nMax,lpfit,lpdx,lpsize);
}
BOOL GetTextExtentExPointW(HDC hDC,LPCTSTR lpsz,int cch,int nMax,LPINT lpfit,LPINT lpdx, LPSIZE lpsize)
{
	char *lpstr;
	lpstr = MakeTempString((LPWSTR)lpsz,0);
	APISTR((LF_API,"GetTextExtentExPointW(.%s..)\n",lpstr));
	return GetTextExtentExPoint(hDC,lpstr,cch,nMax,lpfit,lpdx,lpsize);
}

int
GetWindowLongA(HWND hWnd,int offset)
{
	return GetWindowLong(hWnd,offset);
}

int
GetWindowLongW(HWND hWnd,int offset)
{
	return GetWindowLong(hWnd,offset);
}

int
SetWindowLongA(HWND hWnd,int offset,int value)
{
	return SetWindowLong(hWnd,offset,value);
}
int
SetWindowLongW(HWND hWnd,int offset,int value)
{
	return SetWindowLong(hWnd,offset,value);
}

BOOL SetWindowTextA(HWND hWnd,LPCSTR lpstr)
{
	return SetWindowText(hWnd,lpstr);
}
BOOL
SetWindowTextW(HWND hWnd,LPCSTR lpstr)
{
	return SetWindowText(hWnd,MakeTempString((LPWSTR)lpstr,0));
}


BOOL
AppendMenuA(HMENU hMenu,int flag,int id,LPWSTR *item)
{
	BOOL rc; 
	APISTR((LF_APICALL,"AppendMenuA(%x,%x,%x,%p)\n",
		hMenu,flag,id,item));
	rc = AppendMenu(hMenu,flag,id,(LPCSTR) item);
	APISTR((LF_APIRET,"AppendMenuA returns BOOL %d\n",rc));
	return rc;
}

BOOL
AppendMenuW(HMENU hMenu,int flag,int id,LPWSTR *item)
{
// TODO - is item WIDE???
	APISTR((LF_APICALL,"AppendMenuW(%x,%x,%x,%p)\n",
		hMenu,flag,id,item));
	return 1;
}

LONG
SendDlgItemMessageA(HWND h,int id,UINT msg,WPARAM wparam,LPARAM lparam)
{
	return SendDlgItemMessage(h,id,msg,wparam,lparam);
}

LONG
SendDlgItemMessageW(HWND h,int id,UINT msg,WPARAM wparam,LPARAM lparam)
{
	return SendDlgItemMessage(h,id,msg,wparam,lparam);
}

int   
DrawTextA(HDC hDC, LPCSTR lpsz, int cb, LPRECT lprc, UINT uFormat)
{
	return DrawText(hDC,lpsz,cb,lprc,uFormat);
}

LRESULT
CallWindowProcA(FARPROC lpfnwndproc,HWND hWnd,UINT message,WPARAM w,LPARAM l)
{
	return CallWindowProc(lpfnwndproc,hWnd,message,w,l);	
}

int
SetClassLongA(HWND hWnd, int nIndex, LONG dwNewLong)
{
	return SetClassLong(hWnd,nIndex,dwNewLong);
}
int
GetCurrentDirectoryA(DWORD cchCurDir, LPTSTR lpszCurDir)
{
	return GetCurrentDirectoryA(cchCurDir,lpszCurDir);
}

HWND WINAPI
FindWindowA(LPCSTR lpszClass, LPCSTR lpszWindow)
{
	HWND hWnd;
	APISTR((LF_APICALL,"FindWindowA()\n"));
	hWnd =  FindWindowEx((HWND)NULL, (HWND)NULL, lpszClass, lpszWindow);
	APISTR((LF_APIRET,"FindWindowA returns HWND %x\n",hWnd));
	return hWnd;
}

HWND WINAPI
FindWindowW(LPCSTR lpszClass, LPCSTR lpszWindow)
{
	HWND hWnd;
	APISTR((LF_APICALL,"FindWindowW()\n"));
	hWnd = FindWindowEx((HWND)NULL, (HWND)NULL, lpszClass, lpszWindow);
	APISTR((LF_APIRET,"FindWindowW returns HWND %x\n",hWnd));
	return hWnd;
}

HWND WINAPI
FindWindowExA(HWND hwndParent, HWND hwndChildAfter,
	LPCTSTR lpszClass, LPCTSTR lpszWindow)
{
	HWND hWnd;
	APISTR((LF_APICALL,"FindWindowExA()\n"));
	hWnd = FindWindowEx(hwndParent,hwndChildAfter,lpszClass,lpszWindow);
	APISTR((LF_APIRET,"FindWindowExA returns HWND %x\n",hWnd));
	return hWnd;
}

HWND WINAPI
FindWindowExW(HWND hwndParent, HWND hwndChildAfter,
	LPCTSTR lpszClass, LPCTSTR lpszWindow)
{
	HWND hWnd;
	APISTR((LF_APICALL,"FindWindowExW()\n"));
	hWnd =  FindWindowEx(hwndParent,hwndChildAfter,lpszClass,lpszWindow);
	APISTR((LF_APIRET,"FindWindowExW returns HWND %x\n",hWnd));
	return hWnd;
}

BOOL WINAPI
PeekMessageA(LPMSG lpMsg, HWND hWnd, UINT uMin, UINT uMax,UINT uFlg)
{
	return PeekMessage(lpMsg, hWnd, uMin, uMax,uFlg);
}

BOOL WINAPI
PostMessageA(HWND hWnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{
	return PostMessage(hWnd, wMsg, wParam, lParam);
}

HDC 
CreateMetaFileA(LPCSTR filename)
{
	return CreateMetaFile(filename);
}

HDC
CreateDCA(LPCSTR lpszDriver, LPCSTR lpszDevice,
	 LPCSTR lpszOutput, const void *lpvInitData)
{
	return CreateDC(lpszDriver, lpszDevice, lpszOutput, lpvInitData);
}

void
Shell32End()
{}
void
User32End()
{}
void
Kernel32End()
{}
void
Gdi32End()
{}
