
#include <stdio.h>

#include "windows.h"
#include "Log.h"
#include "winnt.h"
#include "heap.h"
#include "stdarg.h"
#include "Resources.h"
#include "Module.h"

static DWORD freeListSizes[4] = {
	0x20, 0x80, 0x200, 0xffffffff
};

void
HPFreeSub(HANDLE hHeap, void *pHeapRegion, DWORD size, DWORD flags)
{
	APISTR((LF_API,"HPFreeSub\n"));
}

DWORD
HPInit(HANDLE hHeap,void *pHeapRegion,DWORD size,DWORD flags)
{
	HEAP_HEADER pHeap;
	HEAP_ARENA  *pLast;
	HEAP_ARENA  *pArena, *pArena2;
	DWORD       FirstArena;
	FREE_LIST_HEADER *pFreeListEntry, *pFreeListArrayEnd;
	DWORD *pFreeListSize;

	/* ensure that region is aligned on a page boundary */

	pHeap = (HEAP_HEADER) pHeapRegion;

	if (!flags & 0x100) {
	   pHeap->nextHeap = 0;
	   pHeap->nextBlock = 0;
	
	   FirstArena = sizeof(HEAP_HEADER);
	
	   pHeap->signature = 0x4948;
	   pHeap->flags     = flags;
	   pHeap->dwSize      = size;

	   if (!(flags & HEAP_NO_SERIALIZE)) {
		/* InitializeCriticalSection(&pHeap->criticalSection); */	
		/* pHeap->pCriticalSection = ... */
	   }

	   pFreeListArrayEnd = (FREE_LIST_HEADER *) &pHeap->freeListArray + 
			(4 * sizeof(FREE_LIST_HEADER));

	   pFreeListEntry = pHeap->freeListArray;
	
 	   pFreeListSize  = freeListSizes; /* static array of sizes */          

	   while (pFreeListEntry < pFreeListArrayEnd ) {
		pFreeListEntry->dwSize = *pFreeListSize;
		pFreeListEntry->arena.dwSize = 0xa0000001;
		pFreeListEntry->arena.signature = 0x4846;
		pFreeListEntry->arena.prev      = 0; /* what ?*/
		pFreeListEntry->arena.next      = 0; /* what ?*/
		
		pFreeListEntry++;
		pFreeListSize++;
	    }
	
	    /* hook up hte first and last */

	    /* point at first */
	    pArena = &pHeap->freeListArray[0].arena; 

	    /* point at last */
	    pArena2 = &pHeap->freeListArray[3].arena; 

	    pArena->next = pArena2;
	    pArena2->prev = pArena;

	} else {
	
	    FirstArena = 8;

	}

	pHeap->dwSize = size;

	pLast= (HEAP_ARENA *)((int)pHeap + size - 0x10);
	
	/* verify last page committed */

	pLast->dwSize = 0xa0000000;
	pLast->signature = 0x4842;
	
	if ((!(flags & 0x400) && size > 0x1000)) {
	   size -= FirstArena;
	   size -= 0xffc;
	 
	   pArena = pHeapRegion + size + FirstArena;

	   pArena->dwSize = 0xa0000010;
	   pArena->signature = 0x4842;

	   HPFreeSub(hHeap, pArena + 0x10, 0xfdc, 0);

	} else {
	  size -= FirstArena;
	  size -= 0x10;
	}

	HPFreeSub(hHeap, pHeapRegion + FirstArena, size, 0);

	return (DWORD) pHeapRegion;
}

HANDLE
HeapCreate(DWORD options,DWORD isize,DWORD msize)
{
	DWORD rc;
	HANDLE hHeap;
	void  *BaseAddress;

	LOGSTR((LF_APICALL,"HeapCreate(%x,%x,%x)\n",options,isize,msize));

	if (msize && msize < isize) {

		LOGSTR((LF_APIFAIL,"HeapCreate returns DWORD 0\n"));
		SetLastError(ERROR_INVALID_PARAMETER);

		return 0;
	}

	msize += 0xfff;
	msize &= 0xfffff000;

	if (msize == 0) {
		options |= HEAP_FREE_CHECKING_ENABLED;

		isize   &= 0xfffff000;
		msize    = isize +  0x10000; //0x100000; RESERVED...
	}

	BaseAddress = VirtualAlloc(0,msize,MEM_COMMIT,PAGE_EXECUTE_READWRITE);

	hHeap = (HANDLE) BaseAddress; 	

	/* only use these flags when calling HPInit */
	options &= (HEAP_FREE_CHECKING_ENABLED | HEAP_GENERATE_EXCEPTIONS |
		HEAP_NO_SERIALIZE);
	
	rc = HPInit(hHeap,BaseAddress,msize,options);
	 
	LOGSTR((LF_APIRET,"HeapCreate returns DWORD %p\n",rc));
	return rc;
}

int
HeapAlloc(int heap,int flags,int size)
{
	void *BaseAddress;
	APISTR((LF_APICALL,"HeapAlloc(%x,%x,%x)\n",heap,flags,size));
	BaseAddress = VirtualAlloc(0,size,MEM_COMMIT,PAGE_EXECUTE_READWRITE);
	APISTR((LF_APIRET,"HeapAlloc returns int %d\n",(int)BaseAddress));
	return (int) BaseAddress;
}

BOOL
HeapFree(int heap, DWORD dwFlags, LPVOID lpMemory)
{
	APISTR((LF_APICALL,"HeapFree(%x,%x,%p)\n",heap,dwFlags,lpMemory));
	APISTR((LF_APIRET,"HeapFree returns TRUE\n"));
	return TRUE;
}
