/*
 * User Interface Functions
 *
 * Copyright 1997 Dimitrie O. Paun
 */

#include <stdio.h>
#include "windows.h"
#include "Log.h"

/***********************************************************************
 *           UITOOLS_DrawDiagEdge
 *
 * Same as DrawEdge, but with BF_DIAGONAL
 * I tested it extensively and as far as I can tell it is identical to the
 * implementaion in Win95.
 * I do not like that I create and
 * use the 3 Pens to draw the diagonals. It would be better to draw them
 * using the brushes returned by GetSysColorBrush func, but I did not have
 * the patience to implement that yet.
 */
static BOOL UITOOLS_DrawDiagEdge(HDC hdc, RECT *rect, UINT edge,
                                   UINT flags)
{
  HPEN facePen, shadowPen, lightPen, blackPen, grayPen, nullPen;
  HPEN iPen, oPen, oldPen;
  HBRUSH oldBrush, faceBrush;
  int cl, cr, ct, cb;
  BOOL mainDiag;
  POINT tp;
  RECT r;

  /* If both rasied and sunken is specified, they anihilate one another */
  if( !((flags & BF_MONO) || (flags & BF_FLAT)) ){
    if( (edge & BDR_RAISEDOUTER) && (edge & BDR_SUNKENOUTER) )
      return FALSE;
    if( (edge & BDR_RAISEDINNER) && (edge & BDR_SUNKENINNER) )
      return FALSE;
  }

  /* Create/get the tools of the trade... */
  facePen = CreatePen(PS_SOLID, 0, GetSysColor(COLOR_BTNFACE));
  shadowPen = CreatePen(PS_SOLID, 0, GetSysColor(COLOR_BTNSHADOW));
  lightPen = CreatePen(PS_SOLID, 0, GetSysColor(COLOR_BTNHILIGHT));
  grayPen = CreatePen(PS_SOLID, 0, RGB(168, 152, 144));
  blackPen = GetStockObject(BLACK_PEN);
  nullPen = GetStockObject(NULL_PEN);
  faceBrush = GetSysColorBrush(COLOR_BTNFACE);
  oldPen = SelectObject(hdc, nullPen);
  oldBrush = SelectObject(hdc, faceBrush);

  /* this is my working rectangle */
  r = *rect;

  if(flags & BF_MONO){
    oPen = blackPen;
    iPen = nullPen;
  }else if(flags & BF_FLAT){
    oPen = shadowPen;
    iPen = facePen;
  }else {
    if(flags & BF_SOFT){
      if(flags & BF_BOTTOM){
	oPen = (edge & BDR_RAISEDOUTER) ? blackPen : lightPen;
	iPen = (edge & BDR_RAISEDINNER) ? shadowPen : grayPen;
      }
      else{
	oPen = (edge & BDR_RAISEDOUTER) ? lightPen : blackPen;
	iPen = (edge & BDR_RAISEDINNER) ? grayPen : shadowPen;
      }
    }
    else{
      if(flags & BF_BOTTOM){
	oPen = (edge & BDR_RAISEDOUTER) ? blackPen : lightPen;
	iPen = (edge & BDR_RAISEDINNER) ? shadowPen : grayPen;
      }
      else{
	oPen = (edge & BDR_RAISEDOUTER) ? grayPen : shadowPen;
	iPen = (edge & BDR_RAISEDINNER) ? lightPen : blackPen;
      }
    }
  }

  if(flags & BF_BOTTOM){
    if(flags & BF_LEFT){
      cr = -1; cl =  0;
      ct =  0; cb = -1;
      mainDiag = TRUE;
      tp.x = r.left; tp.y = r.top;
    }
    else{ /* RIGHT */
      cr = -1; cl =  0;
      ct =  1; cb =  0;
      tp.x = r.left; tp.y = r.bottom-1;
      mainDiag = FALSE;
    }
  }
  else{  /* TOP */
    if(flags & BF_LEFT){
      cr =  0; cl =  1;
      ct =  0; cb = -1;
      mainDiag = FALSE;
      tp.x = r.right; tp.y = r.top;
    }
    else{ /* RIGHT */
      cr =  0; cl =  1;
      ct =  1; cb =  0;
      tp.x = r.right; tp.y = r.bottom-1;
      mainDiag = TRUE;
    }
  }

  /* if it has external edge, draw it */
  if(edge & BDR_OUTER){
    SelectObject(hdc, oPen);
    MoveToEx(hdc, r.left, mainDiag ? r.bottom-1 : r.top, 0);
    LineTo(hdc, r.right, mainDiag ? r.top-1 : r.bottom);
    r.left += cl; r.right += cr; r.top += ct; r.bottom += cb;
  }

  /* if it has internal edge, draw it */
  if(edge & BDR_INNER){
    SelectObject(hdc, iPen);
    MoveToEx(hdc, r.left, mainDiag ? r.bottom-1 : r.top, 0);
    LineTo(hdc, r.right, mainDiag ? r.top-1 : r.bottom);
    r.left += cl; r.right += cr; r.top += ct; r.bottom += cb;
  }

  if((flags & BF_MIDDLE) && !(flags & BF_MONO)){
    POINT p[3];
    p[0].x = mainDiag ? r.right: r.left;
    p[0].y = r.top;
    p[1].x = mainDiag ? r.left : r.right;
    p[1].y = r.bottom;
    p[2].x = tp.x;
    p[2].y = tp.y;
    SelectObject(hdc, nullPen);
    SelectObject(hdc, faceBrush);
    Polygon(hdc, p, 3);
  }

  if(flags & BF_ADJUST)
    *rect = r;

  /* Restore the DC */
  SelectObject(hdc, oldPen);
  SelectObject(hdc, oldBrush);

  /* Clean-up */
  DeleteObject(facePen);
  DeleteObject(shadowPen);
  DeleteObject(lightPen);
  DeleteObject(grayPen);

  return TRUE;
}


/***********************************************************************
 *           UITOOLS_DrawRectEdge
 *
 * Same as DrawEdge, but without BF_DIAGONAL
 * I tested this function and it works very well. You should not change it
 * unless you find a bug. If you don't like the colors, it it not its 
 * fault - the system colors are not OK.
 * Again, I tested this function on Win95 and I compared the output with the
 * one generated by the native DrawEdge and it is identical on all cases that
 * I tried, and I tried quite a few.
 */
static BOOL UITOOLS_DrawRectEdge(HDC hdc, RECT *rect,
                                   UINT edge, UINT flags)
{
  HBRUSH faceBrush, shadowBrush, lightBrush, blackBrush, grayBrush, nullBrush;
  HBRUSH iNBrush, iSBrush, iEBrush, iWBrush;
  HBRUSH oNBrush, oSBrush, oEBrush, oWBrush;
  HBRUSH oldBrush;
  RECT r;

  /* If both rasied and sunken is specified, they anihilate one another */
  if( !((flags & BF_MONO) || (flags & BF_FLAT)) ){
    if( (edge & BDR_RAISEDOUTER) && (edge & BDR_SUNKENOUTER) )
      return FALSE;
    if( (edge & BDR_RAISEDINNER) && (edge & BDR_SUNKENINNER) )
      return FALSE;
  }

  faceBrush = GetSysColorBrush(COLOR_BTNFACE);
  shadowBrush = GetSysColorBrush(COLOR_BTNSHADOW);
  lightBrush = GetSysColorBrush(COLOR_BTNHILIGHT);
  blackBrush = GetStockObject(BLACK_BRUSH);
  grayBrush = GetStockObject(LTGRAY_BRUSH);
  nullBrush = GetStockObject(NULL_BRUSH);
  oldBrush = SelectObject(hdc, nullBrush);

  /* this is my working rectangle */
  r = *rect;

  if(flags & BF_MONO){
    oNBrush = oSBrush = oEBrush = oWBrush = blackBrush;
    iNBrush = iSBrush = iEBrush = iWBrush = nullBrush;
  }else if(flags & BF_FLAT){
    oNBrush = oSBrush = oEBrush = oWBrush = shadowBrush;
    iNBrush = iSBrush = iEBrush = iWBrush = faceBrush;
  }else 
  {
    if(flags & BF_SOFT){
      oNBrush = oWBrush = (edge & BDR_RAISEDOUTER) ? lightBrush  : blackBrush;
      oSBrush = oEBrush = (edge & BDR_RAISEDOUTER) ? blackBrush  : lightBrush;
      iNBrush = iWBrush = (edge & BDR_RAISEDINNER) ? grayBrush   : shadowBrush;
      iSBrush = iEBrush = (edge & BDR_RAISEDINNER) ? shadowBrush : grayBrush;
    }
    else{
      oNBrush = oWBrush = (edge & BDR_RAISEDOUTER) ? grayBrush   : shadowBrush;
      oSBrush = oEBrush = (edge & BDR_RAISEDOUTER) ? blackBrush  : lightBrush;
      iNBrush = iWBrush = (edge & BDR_RAISEDINNER) ? lightBrush  : blackBrush;
      iSBrush = iEBrush = (edge & BDR_RAISEDINNER) ? shadowBrush : grayBrush;
    }
  }

  /* if it has external edge, draw it */
  if(edge & BDR_OUTER){
    if(flags & BF_RIGHT){
      SelectObject(hdc, oEBrush);
      PatBlt(hdc, r.right-1, r.top, 1, r.bottom - r.top, PATCOPY);
      r.right--;
    }
    if(flags & BF_BOTTOM){
      SelectObject(hdc, oSBrush);
      PatBlt(hdc, r.left, r.bottom-1, r.right-r.left, 1, PATCOPY);
      r.bottom--;
    }
    if(flags & BF_LEFT){
      SelectObject(hdc, oWBrush);
      PatBlt(hdc, r.left, r.top, 1, r.bottom - r.top, PATCOPY);
      r.left++;
    }
    if(flags & BF_TOP){
      SelectObject(hdc, oNBrush);
      PatBlt(hdc, r.left, r.top, r.right-r.left, 1, PATCOPY);
      r.top++;
    }
  }

  /* if it has internal edge, draw it */
  if(edge & BDR_INNER){
    if(flags & BF_RIGHT){
      SelectObject(hdc, iEBrush);
      PatBlt(hdc, r.right-1, r.top, 1, r.bottom - r.top, PATCOPY);
      r.right--;
    }
    if(flags & BF_BOTTOM){
      SelectObject(hdc, iSBrush);
      PatBlt(hdc, r.left, r.bottom-1, r.right-r.left, 1, PATCOPY);
      r.bottom--;
    }
    if(flags & BF_LEFT){
      SelectObject(hdc, iWBrush);
      PatBlt(hdc, r.left, r.top, 1, r.bottom - r.top, PATCOPY);
      r.left++;
    }
    if(flags & BF_TOP){
      SelectObject(hdc, iNBrush);
      PatBlt(hdc, r.left, r.top, r.right-r.left, 1, PATCOPY);
      r.top++;
    }
  }

  /* if we got to fill the middle, to it now */
  if((flags & BF_MIDDLE) && !(flags & BF_MONO))
    FillRect(hdc, &r, faceBrush);

  /* adjust the rectangle if required */
  if(flags & BF_ADJUST)
    *rect = r;

  /* Restore the DC */
  SelectObject(hdc, oldBrush);

  return TRUE;
}


/**********************************************************************
 *          DrawEdge   (USER32.154)
 */
BOOL WINAPI DrawEdge( HDC hdc, LPRECT rc, UINT edge, UINT flags )
{
    BOOL bReturn;
    LOGSTR(( LF_APICALL, "DrawEdge: %04x %d,%d-%d,%d %04x %04x\n",
                      hdc, rc->left, rc->top, rc->right, rc->bottom,
                      edge, flags ));

    if(flags & BF_DIAGONAL)
      bReturn =  UITOOLS_DrawDiagEdge(hdc, rc, edge, flags);
    else
      bReturn =  UITOOLS_DrawRectEdge(hdc, rc, edge, flags);

    LOGSTR(( LF_APIRET, "DrawEdge returns BOOL %d\n",bReturn));
    return bReturn;
}
