/*    
	LoadLibrary.c	2.51
    	Copyright 1997 Willows Software, Inc. 

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; see the file COPYING.LIB.  If
not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA.


For more information about the Willows Twin Libraries.

	http://www.willows.com	

To send email to the maintainer of the Willows Twin Libraries.

	mailto:twin@willows.com 

 */

#include <stdio.h>
#include <string.h>

#include "windows.h"
#include "windowsx.h"

#include "Endian.h"
#include "dos.h"
#include "Log.h"
#include "kerndef.h"
#include "Resources.h"
#include "Module.h"
#include "Kernel.h"
#include "BinTypes.h"
#include "WinConfig.h"
#include "WinDefs.h"
#include <ctype.h>
#include "compat.h"
#include "Driver.h"

extern HMODULE CreateModuleInstance(LPSTR);
extern void FreeModuleInstance(HMODULE);
extern LPSTR strpbrkr(LPCSTR,LPCSTR);
extern BOOL InitBinary(LPSTR,UINT);

extern BOOL InitClasses(void);

/* From LoadDLL.c */
extern int LoadNewInstance(MODULEINFO *, LPSTR, WORD);
extern void LoadModuleFromDscr(MODULEINFO *,LPMODULETAB,WORD);
extern int  LoadModuleFromFile(MODULEINFO *,LPSTR,LPSTR,WORD);

extern BOOL TWIN_InitDriver();

HANDLE TWIN_LoadFile(WORD , LPSTR , LPARAM );
HANDLE TWIN_LoadMDT(WORD , LPSTR , LPARAM );

#ifdef STRICT
typedef FARPROC (CALLBACK* GETPADDRPROC)(LPMODULEINFO, UINT);
#else
typedef FARPROC GETPADDRPROC;
#endif

GETPADDRPROC lpfnOEMGetPAddr = NULL;

#ifdef TWIN32
DWORD          GetModuleFileName(HMODULE,LPTSTR,DWORD);
#else
int          GetModuleFileName(HINSTANCE,LPSTR,int);
#endif
FARPROC      GetProcAddress(HINSTANCE,LPCSTR);
HINSTANCE    InternalLoadDLL(MODULETAB *);
static DWORD  InternalLoadNativeDLL(LPCSTR);


extern LPOBJHEAD lpModuleHead;
extern MODULETAB TWIN_ModuleTable[];

HINSTANCE
LoadLibrary(LPCSTR lpszLibFileName)
{
    HANDLE hModule = 0;
    HINSTANCE hInst;

    APISTR((LF_APICALL,"LoadLibrary(%s)\n",
	lpszLibFileName?lpszLibFileName:"NULL"));

    if (!strlen(lpszLibFileName)) {
	APISTR((LF_APIFAIL,"LoadLibrary: returns HINSTANCE %x\n",0)); 
	return (HINSTANCE)0;
    }

    /* try to find the .dll that corresponds with the library */
    hModule = TWIN_LoadFile(ILL_NOERROR|ILL_NEWINST,"",
			      (LPARAM)lpszLibFileName);

    /* what is the data instance we will use for it? */
    hInst = GetInstanceFromModule(hModule);

    APISTR((LF_APIRET,"LoadLibrary: returns HINSTANCE %x\n",hInst)); 
    return hInst;
}

HINSTANCE
InternalLoadDLL(MODULETAB *lpModTable)
{
	int n;
	static int bLoadImmediate;

	TWIN_InitDriver();

	if(lpModTable == 0) {
		bLoadImmediate++;
		return 0;
	}
	if(bLoadImmediate) {
        	HMODULE hModule;
		HINSTANCE hInstance;

    		hModule = TWIN_LoadMDT(ILL_NEWINST|ILL_DSCR,"",(LPARAM)lpModTable);
		hInstance = GetInstanceFromModule(hModule);

		return hInstance;
	}

	for(n=1;n<256;n++) {

		if(TWIN_ModuleTable[n].flags == 0) {

			TWIN_ModuleTable[n].flags = lpModTable->flags;
			TWIN_ModuleTable[n].dscr = lpModTable->dscr;
			break;
		}
	}

	return 0;
}

static DWORD
InternalLoadNativeDLL(LPCSTR lpszLibFileName)
{
	DWORD	hSO;

	hSO =  DRVCALL_SYSTEM(PSSH_LOADLIBRARY,0,0,lpszLibFileName);

	return hSO;
}

void WINAPI
FreeLibrary(HINSTANCE hInst)
{
    HMODULE hModule;
    MODULEINFO *modinfo;

    APISTR((LF_APICALL,"FreeLibrary(HINSTANCE=%x)\n",hInst));

    if (!(modinfo = GETMODULEINFO(hInst))) {
	hModule = GetModuleFromInstance(hInst);
	if (!(modinfo = GETMODULEINFO(hModule))) {
	    APISTR((LF_APIFAIL,"FreeLibrary: returns void\n"));
	    return;
	}
    }

#ifdef LATER
    decrement reference count, if NULL -- free module
#endif

    RELEASEMODULEINFO(modinfo);

    APISTR((LF_APIRET,"FreeLibrary: returns void\n"));
}

HANDLE
TWIN_LoadFile(WORD flags, LPSTR lpCmdLine, LPARAM lParam)
{
        LPMODULETAB mod_tab = 0,mdt;
        LPMODULEDSCR mod_dscr = 0;
        char szModuleName[80];
        char FileName[_MAX_PATH];
        LPSTR lpszFileName = 0, lpszModuleName;
        HMODULE hModule;
        LPMODULEINFO modinfo = 0;
        DWORD hSO;
	DWORD ret;

	/* we have a filename... */
	lpszFileName = (LPSTR)lParam;

	/* get just the component part... */
	_splitpath(lpszFileName,NULL,NULL,szModuleName,NULL);

        /* special case, this is a builtin module... */
        /* internally, it is called KERNEL, but this is the same... */
    
        if (!strcasecmp(szModuleName,"KRNL386")) {	/* KERNEL on 386 CPU */
	   return GetModuleHandle("KERNEL");
        }

	/********************************************************/
	/*  Step 1. Check if already loaded.			*/
	/********************************************************/

	hModule = GetModuleHandle(szModuleName);

	if(hModule) {
		return hModule;
	}

	/********************************************************/
	/*  Step 2. Alias substitution				*/
	/********************************************************/

	if ((GetPrivateProfileString("boot",lpszFileName,"",FileName,
		sizeof(FileName), GetTwinFilename())))
	    lpszFileName = FileName;

	/********************************************************/
	/*  Step 3. Try to load a native version of the file.  	*/ 
	/********************************************************/

	if ((hSO = InternalLoadNativeDLL(lpszFileName))) {
                
		/* we loaded a native library, so return its handle */
		hModule = GetModuleHandle(szModuleName);
		if(hModule == 0) {

			hModule = CreateModuleInstance(szModuleName);
			modinfo = GETMODULEINFO(hModule);
			modinfo->lpFileName = lpCmdLine;
		}

		return hModule;
	} 

	/********************************************************/
	/* Step 3. See if the module has already been loaded    */
	/* 	   This can be binary or native			*/
	/********************************************************/
	   
	if ((hModule = GetModuleHandle(szModuleName))) {
	    modinfo = GETMODULEINFO(hModule);

	    /* If modinfo->wSelBase is not set => */
	    /* it means that we returned to already loaded module to */
	    /* support.					  	     */

	    if (!(!modinfo->wSelBase ) &&
		(!(flags & ILL_NEWINST) || !modinfo->fMakeInstance))
	    {
	        RELEASEMODULEINFO(modinfo);
		return hModule;
	    }
	}

	/********************************************************/
    	/* Step 4. Module alias check.				*/
    	/*     e.g COMMDLG.DRV = COMMGLG.DRV			*/
	/*							*/
	/* The purpose of this code is to allow real binary     */
	/* dll's to replace our remaining internal dll's. 	*/
	/* 							*/
	/* Use the above alias to use binary vs. native dll's   */
	/* Use this to use binary vs. built-in dll's		*/
	/********************************************************/

	if (GetPrivateProfileString("boot",szModuleName,"",FileName,
			sizeof(FileName), GetTwinFilename())) 
	    lpszFileName = FileName;

	else { /* Try to find a descriptor in mdt */
	    int j = 0;

	    mdt = &TWIN_ModuleTable[0];
	    while(mdt[++j].dscr)
		if (!strcasecmp(mdt[j].dscr->name,szModuleName)) {
		    mod_dscr = mdt[j].dscr;
		    mod_tab = &mdt[j];
		    break;
		}
	}

        if (!mod_dscr) {

            /* we don't have a module descriptor, so we must have a file? */
	    InitBinary(lpCmdLine,1); 

        }


	/********************************************************/
        /* Create a hModule for this file 			*/
	/********************************************************/

        if (!hModule) {
	    hModule = CreateModuleInstance(szModuleName);
	    modinfo = GETMODULEINFO(hModule);

	    /* set the filename */
	    modinfo->lpFileName = lpCmdLine;
        }

	/********************************************************/
        /* we have a module descriptor 				*/
	/********************************************************/

	lpszModuleName = szModuleName;

        if (modinfo && modinfo->fMakeInstance) {
	    LoadNewInstance(modinfo, lpCmdLine, flags);
        }
        else if (mod_dscr) {
    
	    LoadModuleFromDscr(modinfo,mod_tab,flags);

        } else {	/* no dscr, just name */

	    if (!(ret = LoadModuleFromFile(modinfo,lpszFileName,lpCmdLine,flags)))
	       FreeModuleInstance(hModule);

	    RELEASEMODULEINFO(modinfo);
	 
	    return (HMODULE)ret;
        }
    
	RELEASEMODULEINFO(modinfo);
    	return hModule;
}

HANDLE
TWIN_LoadMDT(WORD flags, LPSTR lpCmdLine, LPARAM lParam)
{
        LPMODULETAB mod_tab;
        LPMODULEDSCR mod_dscr;
        LPSTR lpszModuleName;
        HMODULE hModule;
        LPMODULEINFO modinfo;

	/* we have a module descriptor table */
	mod_tab = (LPMODULETAB)lParam;
	mod_dscr = (LPMODULEDSCR)mod_tab->dscr;

	/* what is the module name ? */
	lpszModuleName = mod_dscr->name;

	/* if its loaded already, return its hModule */
	if ((hModule = GetModuleHandle(lpszModuleName))) {
	    return hModule;
	}

	/* not loaded so create the hModule for it */
	hModule = CreateModuleInstance(lpszModuleName);
	modinfo = GETMODULEINFO(hModule);

	/* set the filename */
	modinfo->lpFileName = lpCmdLine;

    	if (modinfo->fMakeInstance) {
		LoadNewInstance(modinfo, lpCmdLine, flags);
    	} else {
		LoadModuleFromDscr(modinfo,mod_tab,flags);
        }
    
	RELEASEMODULEINFO(modinfo);
	return hModule;
}

static DWORD
TWIN_GetModuleFileName(HMODULE hInstance,LPTSTR lpszFileName, DWORD cbFileName,int flag)
{
    HMODULE hModule;
    LPMODULEINFO lpModuleInfo;
    int len;

    if (!(hModule = GetModuleFromInstance(hInstance)))
	hModule = hInstance;
    if (!(lpModuleInfo = GETMODULEINFO(hModule)))
	return 0;

    if (lpModuleInfo->lpFileName) {
	len = strlen(strncpy(lpszFileName,lpModuleInfo->lpFileName,cbFileName));
    } else {

//	if (!strcasecmp(lpModuleInfo->lpModuleName,"COMMDLG")) {
//	    len = strlen(strncpy(lpszFileName,"commdlg.dll",cbFileName));
//	} else 
	{
	    GetCurrentDirectory(cbFileName,lpszFileName);
	    strcat(lpszFileName,"/");
	    strcat(lpszFileName,lpModuleInfo->lpModuleName);
	    len = strlen(lpszFileName);
	}
    }
    RELEASEMODULEINFO(lpModuleInfo);
    return len;
}
#ifdef TWIN32
DWORD
GetModuleFileName(HMODULE hInstance,LPTSTR lpszFileName, DWORD cbFileName)
#else
int
GetModuleFileName(HINSTANCE hInstance,LPSTR lpszFileName, int cbFileName)
#endif
{
	return TWIN_GetModuleFileName(hInstance,lpszFileName, cbFileName,1);
}

FARPROC WINAPI
GetProcAddress(HINSTANCE hInstance, LPCSTR lpszProc)
{
    HANDLE hModule;
    LPMODULEINFO lpModuleInfo;
    ENTRYTAB *lpEntry;
    int i;
    FARPROC retFunc = 0;

    if(hInstance == 0)
       hInstance = GetInstanceFromModule(GetModuleHandle("USER"));

    if (!(lpModuleInfo = GETMODULEINFO(hInstance))) {
	if (!(hModule = GetModuleFromInstance(hInstance)))
	    return (FARPROC)NULL;
	if (!(lpModuleInfo = GETMODULEINFO(hModule)))
	    return (FARPROC)NULL;
    }
    if (!HIWORD((DWORD)lpszProc)) {
	i = LOWORD((DWORD)lpszProc);
	lpEntry = &lpModuleInfo->lpEntryTable[i];
	retFunc =  (FARPROC)lpEntry->fun;
	RELEASEMODULEINFO(lpModuleInfo);
	if (!retFunc && lpfnOEMGetPAddr)
	    return (FARPROC)lpfnOEMGetPAddr(lpModuleInfo,i);
	else
	    return retFunc;
    }
    else {	/* search by name */
	for(i=0,lpEntry = lpModuleInfo->lpEntryTable;
	    lpEntry->name;
	    i++,lpEntry++) 
	    if (!strcasecmp(lpszProc,lpEntry->name)) { 
		    retFunc =  (FARPROC)lpEntry->fun;
		    RELEASEMODULEINFO(lpModuleInfo);
		    if (!retFunc && lpfnOEMGetPAddr)
			return (FARPROC)lpfnOEMGetPAddr(lpModuleInfo,i);
		    else
			return retFunc;
	    }
    }
    RELEASEMODULEINFO(lpModuleInfo);
    return NULL;
}

HMODULE
GetModuleHandle(LPCSTR lpszModule)
{
    MODULEINFO *mi;
    ATOM atom;
    LPOBJHEAD lpModuleTemp;
    LPSTR lp;
    LPSTR lpModName;
    LPSTR lpStr;

    if (!HIWORD(lpszModule)) {
	return GetModuleFromInstance((HINSTANCE)LOWORD((DWORD)lpszModule));
    }

    /* get a copy of the string, we are going to modify it */
    lpStr = WinMalloc(strlen(lpszModule)+1);
    strcpy(lpStr,lpszModule);

    /* extract the filename portion */
    if ((lpModName = strrchr(lpStr,'/'))) {
	lpModName++;
    } else
	lpModName = lpStr;

    /* drop off any extensions */
    if ((lp = strchr(lpModName,'.')))	/* strip off possible extension */
	*lp = '\0';

    if (!(atom = FindAtom(lpModName))) {
	WinFree(lpStr);
	return 0;
    }

    /* lpModName points to our internal copy */

    for (lpModuleTemp = lpModuleHead; lpModuleTemp;
	lpModuleTemp = (LPOBJHEAD)lpModuleTemp->lpObjNext) {
	mi = (LPMODULEINFO)lpModuleTemp;

	if (mi->atmModuleName == atom) {
	    WinFree(lpStr);
	    return lpModuleTemp->hObj;
        }
    }
    /* can nolonger use the string */
    WinFree(lpStr);
    return 0;
}

HINSTANCE 
LoadModule(LPCSTR lpszModuleName, LPVOID lpvParamBlock)
{
    HMODULE hModule = 0;
    PARAMBLOCK *lpb;

    lpb = (PARAMBLOCK *)lpvParamBlock;

    hModule = TWIN_LoadFile(ILL_NOERROR|ILL_EXEC|ILL_NEWINST,
				  lpb->lpszCmdLine,
				  (LPARAM)lpszModuleName);
    return GetInstanceFromModule(hModule);
}

HDRVR
GetNextDriver(HDRVR hDrvr, DWORD fdwFlags)
{
    APISTR((LF_APISTUB,"GetNextDriver(HDRVR=%x,DWORD=%x)\n",
	hDrvr,fdwFlags));
    return 0;
}

HINSTANCE
LoadDriver(LPCSTR lpszDriver)
{
	char buf[_MAX_PATH];
	int i;
 
	for (i=0;lpszDriver[i];i++)
		buf[i] = tolower(lpszDriver[i]);

	buf[i] = '\0';
	strcat(buf,".drv");

	return LoadLibrary(buf);
}

void
TWIN_AppendExtension(char *lpstr,char *lpext)
{
	char *basename = lpstr;
	char *lpbase;

	if ( (lpbase = strrchr(basename,'/')) )
		basename = lpbase + 1;
	if ( (lpbase = strrchr(basename,'\\')) )
		basename = lpbase + 1;
	
	if ( (lpbase = strrchr(basename,'.')) == 0 ) {
	   strcat(lpstr,".");
	} else
	   return;
	
	strcat(lpstr,lpext);
}


UINT WINAPI
WinExec(LPCSTR lpCmdLine, UINT fuCmdShow)
{
	LPSTR lpszProgramName,lpTemp;
	PARAMBLOCK pb;
	char buf[BUFFER];
	UINT uiShow;
	HMODULE hModule;

	APISTR((LF_APICALL,"WinExec(cmdline=%s,cmdshow=%x\n",
		lpCmdLine,fuCmdShow));

	strcpy(buf,lpCmdLine);
	if ((lpTemp = strchr(buf,' '))) {
	    *lpTemp = '\0';
	    pb.lpszCmdLine = (char *) WinStrdup(++lpTemp);
	}
	else pb.lpszCmdLine = (char *) WinStrdup("");

	uiShow = fuCmdShow;
	pb.lpShow = &uiShow;
	pb.wSegEnv = 0;

	TWIN_AppendExtension(buf,"exe");

	lpszProgramName = WinMalloc(strlen(buf)+1);
	strcpy(lpszProgramName,buf);
	hModule = LoadModule(lpszProgramName,(LPVOID)&pb);
	WinFree(lpszProgramName);
	APISTR((LF_APIRET,"WinExec: returns UINT %x\n",hModule));
	return (UINT)hModule;
}

/* GetModuleDescription -- convenience function returning module description */
/* string read from the non-resident name table of the binary header */

LPSTR
GetModuleDescription(HINSTANCE hInstance)
{
    HMODULE hModule;
    LPMODULEINFO lpModuleInfo;
    LPSTR lpszResult;

    if (!(hModule = GetModuleFromInstance(hInstance)))
	hModule = hInstance;
    if (!(lpModuleInfo = GETMODULEINFO(hModule)))
	return 0;
    lpszResult = lpModuleInfo->lpModuleDescr;
    RELEASEMODULEINFO(lpModuleInfo);
    return lpszResult;
}


int
TWIN_GetWin32Flag(HWND hWnd)
{
	HINSTANCE 	hInst;
    	HMODULE 	hModule;
    	LPMODULEINFO 	lpModuleInfo;
	
	hInst = GetWindowInstance(hWnd);	
	hModule = GetModuleFromInstance(hInst);
	lpModuleInfo = GETMODULEINFO(hModule);

	if(lpModuleInfo && lpModuleInfo->dwWin32flag)
		return 1;
	return 0;
}
