
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <ctype.h>
#include "Log.h"
#include "mfs_config.h"

#include "Profile.h"

/************************************************************
   TODO
   some functionality such as
  	delete section does not exist
  		section give, but no key and/or value
    
   on exit, use the full name, and modify that to write to the
   file...  Note: use the path of the original file... ie. the
   fileinfo block HAS an ofs struct in it?  With the original
   path.  Also, fixup the end to also use OpenFile/_lwrite/_lclose,
   and the MFS_RENAME, MFS_DELETE functions...

   some optimization such as
    !!! not writing values that don't change
    !!! setting dirty bit for write functions ONLY when data changes...
   note: even write file functions open the original in readonly mode
   note: don't COPY the string data, but keep a pointer to it, and DO NOT
         write into the buffer, we want the = sign left there to write out
	 the entry when we are done...
      
  
   some thread safe locking
    making sure only one thread accesses data

*************************************************************/

/* linked list of files that we have opened 	*/
/* to read/write profile entries		*/
typedef struct profile_struct {
	struct profile_struct *next;		/* next structure    */
	char *filename;				/* basename of file  */
	int   flag;				/* is it writable?   */
	SEGMENT *sp;				/* start of buffer info   */
	OFSTRUCT ofs;				/* ofs structure for file */
} PROFILE_ENTRY;

static SEGMENT * TWIN_GetNewObject(int ,char *);

/* in-core files */
static PROFILE_ENTRY *inifiles;

/* default filename for profile functions */
#define SZPROFILE	"win.ini"

/* private function to remove spaces around a string... */
static char *
strtrim( char * str)
{
    char *s = str;
    int       slen;

    if (s) {
        while (*s  &&  isspace(*s))
            ++s;
        for (slen=strlen(s); slen>=0
               &&  isspace(*(s-1+slen)); (slen)--);
           *(s + slen) = 0;
    }

    return s;
}

/* private function to read section buffer for a line of data...  */

static char *
readline(SEGMENT *sp,char *lpstr,int size)
{
	char *lp;
	char *np;
	char *mp;
	int   n;

	if(sp->offset == 0)
		sp->offset = sp->szentry;
	
	lp = sp->offset;
	mp = sp->szentry + sp->entrysz;

	if(lp >= mp)
		return sp->offset = 0;

	/* skip over leading newlines... */
	while (*lp && (*lp == '\n' || *lp == '\r'))
		lp++;

	/* the start of this entry */
	sp->szlast = sp->offset;

	if(*lp == 0)
		lp = mp;

	if(lp >= mp)
		return sp->offset = 0;

	/* now find next newline */
	np = lp++;
	while (*lp && (*lp != '\n' && *lp != '\r')) {
		if(*lp == 26)
			break;
		lp++;
		if(lp >= mp)
			break;
	}

	if(lp > mp)
		return sp->offset = 0;

	n = lp - np;

	if(n > size - 1)
		n = size - 1;
	
	while (n && (*np == ' ' || *np == '\t')) {
	      np++;
	      n--;
	}
	strncpy(lpstr,np,n);
	lpstr[n] = 0;

	sp->offset = lp;

	return lpstr;
}

static SEGMENT *
TWIN_GetNewObject(int size,char *sname)
{
	SEGMENT *fp;

	fp = (SEGMENT *) GetFreeObject(SEGSIZE);

	/* allocate buffer for text */
	fp->szentry = (char *) WinMalloc(size + 1);
	memset(fp->szentry,0,size + 1);

	/* fill in section header information */
	fp->entrysz = size;
	fp->next  = 0;
	fp->offset = 0;
	fp->sname = WinStrdup(sname);
	fp->sflag = 1;

	return fp;
}


/* break the .ini file data into sections
 * one for each section
 * comments will be attached to the last section
 * or be in their own initial section
 */
		
void
breaklines(SEGMENT *sp)
{
	char buffer[512];
	char *lp = 0;
	SEGMENT *np;

	while ((lp = readline(sp,buffer,80))) {

		while (*lp == ' ' || *lp == '\t') {
		       lp++;
		}

		/* this is a comment, go get another 		*/
		/* TBD what about newlines, with no data at all */
		/*     skip it internally? 			*/
		/*     or is it okay that it is just data... 	*/
		if (*lp == ';' || *lp == '\0')
			continue;

		/* break this section header to make a new section */
		if (*lp == '[') {
			np = (SEGMENT *)GetFreeObject(SEGSIZE);
		
			/* initialize new entry */
			np->sflag = 0;		   /* non-allocated space */
			np->szentry = sp->offset;  
			np->offset  = 0;
			np->next    = sp->next;	   /* what is after this one */
			np->sname   = WinStrdup(buffer);  /* section name    */
			np->entrysz = sp->entrysz; /* section size, parent   */

			/* this entry now points to new entry */
			sp->next    = np;

			/* this entry ends where this one left off */
			sp->entrysz = sp->szlast - sp->szentry;
			sp->offset  = 0; /* we will read from start next */
			sp = np;
		}
	}
}

int
InitIniFile(char *filename)
{
	struct stat     Status;
	FILE *filep;
	SEGMENT *fp;

	if(inifiles)
		return 0;

	inifiles = (PROFILE_ENTRY *) WinMalloc(sizeof(PROFILE_ENTRY));
	memset(inifiles,0,sizeof(PROFILE_ENTRY));

	inifiles->filename = WinStrdup(filename);
	inifiles->next = 0;
	inifiles->sp   = 0;
	inifiles->flag = 0;
	strcpy(inifiles->ofs.szPathName,filename);

	/* have not seen it, open anew */
        filep = fopen(filename,"r");
	
	/* if we have found it... */
	if(filep) {
		int i;
		/* how big is the file */
		i = fstat ( fileno(filep), &Status );
		/*printf("file(%s) size=%ld %i\n",filename,Status.st_size,i);*/
		
		/* create a section header for it       */
		/* and allocate space for the object... */
		fp = TWIN_GetNewObject(Status.st_size,"unamed");
		inifiles->sp   = fp;

		/* read the .ini file data */
    		fread(fp->szentry,fp->entrysz,1,filep);

		/* closeup the file */
		fclose(filep);

		breaklines(fp);
	}
	return 0;
}

/* private function to open and load a .ini file 		*/
/* file == 0, then use default file		 		*/
/* mode       mode to open the file, either "r" or "w+"         */
/* flag		  flag for .ini file, ie. readonly/readwrite    */

extern char *GetTwinFilename();

SEGMENT *
LoadIniFile(LPCTSTR file,int flag)
{
	struct stat     Status;
	SEGMENT	*fp = 0;
        HFILE hFile;

	PROFILE_ENTRY *next;
	
	/*printf("LoadIniFile: %s\n",file);*/
	/* use our built-in default name if none give */
	if(file == 0)
		file = SZPROFILE;

	/* make sure preferences file is loaded... */
	InitIniFile(GetTwinFilename());

	/* have we seen this one before */
	for(next = inifiles;next;next = next->next) {
		if((next->filename && strcmp(next->filename,file) == 0)) {
			/* if this is a writable entry... */
			if(flag)
				next->flag = 1;
			if(next->sp)
				(next->sp)->offset = 0;
			return next->sp;
		}
	}

	/* save a linked list of all the files we have loaded
	 * later, we could...
	 *		keep track of last time accessed
	 *		keep track of if tree is dirty
	 *		keep track of how many are open and/or dirty
	 */
	{ 
		PROFILE_ENTRY *next = inifiles;;
			
		inifiles = (PROFILE_ENTRY *) WinMalloc(sizeof(PROFILE_ENTRY));
		memset(inifiles,0,sizeof(PROFILE_ENTRY));

		inifiles->filename = WinStrdup(file);
		inifiles->next = next;
		inifiles->sp   = 0;
		inifiles->flag = flag;
	}

	/* have not seen it, open anew */
        hFile = OpenFile(file,&inifiles->ofs,OF_READ);
	
	/* if we have found it... */
	if(hFile < 0) {
		/* the file does not exist */
		/* create an empty entry for it */

		if(file[0] != '/') {
			GetWindowsDirectory(inifiles->ofs.szPathName,255);	
			strcat(inifiles->ofs.szPathName,"/");
		}
		/* copy the filename... */
		strcat(inifiles->ofs.szPathName,file);

		fp = TWIN_GetNewObject(0,"unamed");

		inifiles->sp   = fp;

	} else {
		int i;
		/* how big is the file */
		/*i = MFS_STAT ( inifiles->ofs.szPathName, &Status );*/
		i = MFS_STAT ( file, &Status );
		/*printf("file(%s) size=%ld %i\n",file,Status.st_size,i);*/
		i = MFS_STAT ( inifiles->ofs.szPathName, &Status );
		/*printf("file(%s) size=%ld %i\n",inifiles->ofs.szPathName,Status.st_size,i);*/
		
		/* create a section header for it */
		/* and allocate space for the object...*/
		fp = TWIN_GetNewObject(Status.st_size,"unamed");
		inifiles->sp   = fp;

		/* read the .ini file data */
    		_lread(hFile,fp->szentry,fp->entrysz);

		/* closeup the file */
		_lclose(hFile);

		breaklines(fp);
	}

	/* here is the file section header */
	return fp;
}

static char *
getlines(SEGMENT **cp,LPCSTR section,char *buffer,int cnt)
{
	SEGMENT *sp = *cp;
	char *lp;

	while ((lp = readline(sp,buffer,cnt))) {

		/* if we have a section name, and it matches */
		if (section) {
			/* then return the string we read */
			if (strnicmp(section,sp->sname+1,strlen(section)) == 0){
				while (*lp == ' ' || *lp == '\t')
				     lp++;
				return lp;
			}
			continue;
		}
		/* we can/will return the string just read */
		return lp;
	}

	/* ran out of the previous section */
	sp = sp->next;

	/* if we have a new section */
	while (sp) {
		SEGMENT *pp = *cp;
		sp->offset = 0;
		*cp = sp;

		/* if this one is different return it, AND we  */
		/* are not trying to read a specific section   */
		if (stricmp(sp->sname,pp->sname) && section == 0) { 
			strncpy(buffer,sp->sname,cnt);
			return buffer;
		}

		/* go ahead and continue reading */
		/* new section, with same name   */
		while((lp = readline(sp,buffer,cnt))) {
			/* if we have a section name, and it matches */
			if (section) {
			/* then return the string we read */
			if (strnicmp(section,sp->sname+1,strlen(section)) == 0){
				while (*lp == ' ' || *lp == '\t')
				     lp++;
					return lp;
				continue;
			}
			}
			/* we can/will return the string just read */
			return lp;
		}
		sp = sp->next;
	}
	
	return 0;
}


/* private function to copy profile string data */
/* data starts at first non-blank after = sign  */
/* TBD should we also skip tabs?		*/
static int
copyentry(char *buffer,int count)
{
	char *lp;

	/* find the value portion of string */
	lp = strchr(buffer,'=');

	/* if we have it  */
	if(lp) { lp++;
		strncpy(buffer,strtrim(lp),count);
	} else
		buffer[0] = 0;

	return strlen(buffer);
}

/*********************************************************************/

/* the actual function for reading from a specific file */

DWORD	WINAPI 
GetPrivateProfileSection(LPCSTR section, LPSTR buffer, DWORD count, LPCSTR file)
{
	SEGMENT *sp;
	char *lp;
	char *bp;
	int   cnt;
	int   n;
	char  datum[512];
	
	/* find section header for file */
	sp = LoadIniFile(file,0);

	/* for section header, search for section */
	while (sp) {
		if(sp->sname[0] == '[' && 
			strnicmp(sp->sname+1,section,strlen(section)) == 0) {

			/*printf("section %s\n",section);*/
			bp = buffer;	/* where to write in buffer */
			cnt = count;	/* remaining space in buffer */
			buffer[0] = 0;	/* build in double terminator */
			buffer[1] = 0;

			sp->offset = 0;	/* guarantee to read from beginning */
			while ((lp = getlines(&sp,section,datum,512))) {
				if (datum[0] == '#')
				   continue;

				lp = strchr(datum,'=');		/* find terminating = sign */
				if (lp)
					*lp = 0;  /* terminate the string */

				lp = datum;
				/* we have no more room for strings, quit */
				if (cnt <= strlen(lp))
				        break;

				/* copy this string into the buffer */
				n = min(cnt,(int)strlen(lp));	/* maximum to copy */
				strncpy(bp,lp,n);  /* copy it */
				bp += n;	   /* update pointer */
				cnt -= n;	   /* decrement remaining space */
				*++bp = 0;	   /* advance to next token */
				cnt--;		   /* additional blank */
			}
			
			return count - cnt;
		}

		/* try next section */
		sp = sp->next;
	}

	return 0;
}

int
GetProfileSection(char *section,char *buffer, int count)
{
	return GetPrivateProfileSection(section,buffer,count,0);
}

static int 
InternalGetProfile(LPCSTR section, LPCSTR entry, LPCSTR lpstr, LPSTR buffer, int count, LPCSTR file)
{
	SEGMENT *sp;
	char *lp;

	/* do we want the whole section, or just the specific keys... */
	if (entry == NULL) {
		return GetPrivateProfileSection(section,buffer,count,file);
	}

	/* find section header for file */
	sp = LoadIniFile(file,0);

	/* for section header, search for section */
	while (sp) {
		if(strnicmp(sp->sname+1,section,strlen(section)) == 0) {

			/* found section, copy section data */
			sp->offset = 0;
			while ((lp = getlines(&sp,section,buffer,count))) {

				if(strnicmp(entry,buffer,strlen(entry)) == 0) {
					return copyentry(buffer,count);
				}
			}
			break;
		}
		if(sp)
			sp = sp->next;
	}

	/* default case */
	if(lpstr)
		strncpy(buffer,lpstr,count);
	else
		strcpy(buffer,"");
	return strlen(buffer);
}

/* the actual function for reading from a specific file */
int     WINAPI 
GetPrivateProfileString(LPCSTR section, LPCSTR entry, LPCSTR lpstr, LPSTR buffer, int count, LPCSTR file)
{
	int rc;
    	APISTR((LF_APICALL, 
	    "GetPrivateProfileString(LPCSTR=%s,LPCSTR=%s,LPCSTR=%s,LPSTR=%x,int %d,LPCSTR=%s)\n",
		section?section:"NULL",
		entry?entry:"NULL",
		lpstr?lpstr:"NULL",
		buffer,
		count,
		(file)?file:"NULL"));
	rc = InternalGetProfile(section,entry,lpstr,buffer,count,file);
        APISTR((LF_APIRET,"GetPrivateProfileString: returns int %d\n",rc));
	return rc;
}

/* read from default .ini file */
int     WINAPI 
GetProfileString(LPCSTR section, LPCSTR entry, LPCSTR lpstr, LPSTR buffer, int count)
{
    int rc;
    APISTR((LF_APICALL,
	"GetProfileString(LPCSTR=%s,LPCSTR=%s,LPCSTR=%s,LPSTR=%x,int %d)\n",
	(section)?section:"NULL",
	(entry)?entry:"NULL",
	(lpstr)?lpstr:"NULL",
	buffer,
	count));
    rc = InternalGetProfile(section,entry,lpstr,buffer,count,0);
    APISTR((LF_APIRET,"GetProfileString: returns int %d\n",rc));
    return rc;
}

/* read integer from private .ini file */
UINT    WINAPI 
GetPrivateProfileInt(LPCSTR section, LPCSTR entry, int rc, LPCSTR file)
{
	char ascii[32];

	APISTR((LF_APICALL, 
	 "GetPrivateProfileInt(LPCSTR=%s,LPCSTR=%s,int=%d)\n", 
		(section)?section:"NULL", 
		(entry)?entry:"NULL", 
		rc));
    
	if (InternalGetProfile(section,entry,0,ascii,32,file))
		rc = strtol(ascii,0,0);
    	APISTR((LF_APIRET,"GetPrivateProfileInt: returning %d\n",rc));
	return rc;
}

/* read integer from default .ini file */
UINT    WINAPI 
GetProfileInt(LPCSTR section, LPCSTR entry , int rc)
{
	char ascii[32];

	APISTR((LF_APICALL, 
	 "GetProfileInt(LPCSTR=%s,LPCSTR=%s,int=%d)\n", 
		(section)?section:"NULL", 
		(entry)?entry:"NULL", 
		rc));
    
	if (InternalGetProfile(section,entry,0,ascii,32,0))
		rc = strtol(ascii,0,0);
    	APISTR((LF_APIRET,"GetProfileInt: returning %d\n",rc));
	return rc;
}

static int
FlushIniCache(LPCSTR filename)
{
	FILE *filep;
	char *lp;
	char  buffer[512];
	char  newfile[128];
	SEGMENT *sp;

	PROFILE_ENTRY *next;
	
	/* have we seen this one before */
	for(next = inifiles;next;next = next->next) {
	
		/* if we have a filename, and it doesn't match */
		/* skip flushing this file... */
		if(filename && stricmp(filename, next->filename))
			continue;

		/* is this file a readwrite ini file, if so it may */
		/* have been written */
		if(next->flag == 1 && next->sp) {

			strcpy(newfile,next->ofs.szPathName);
			if((lp = strrchr(newfile,'.')))
				*lp = 0;
			strcat(newfile,".tmp");
			filep = fopen(newfile,"w+");
			if(filep) {
				sp = next->sp;
				while((lp = getlines(&sp,0,buffer,512))) {
					strcat(buffer,"\n");
					fwrite(buffer,strlen(buffer),1,filep);
					}
				fclose(filep);
				    
				/* remove the previous file */
				unlink(next->ofs.szPathName);

				/* rename the temp file to the original file */
				rename(newfile,next->ofs.szPathName);
			}
		}
	}

	return 0;
}

SEGMENT *
addsection(SEGMENT *sp,int flag,LPCSTR key,LPCSTR value)
{
	SEGMENT *np,*tp;
	char    *lpstr;

	np = (SEGMENT *)GetFreeObject(SEGSIZE);

	/* initialize new entry */
	np->sflag = 1;				/* allocated space */
	if(key == 0)
		key = "";
	lpstr = WinMalloc(strlen(key) + strlen(value?value:"") + 4); /* strings, = and null */

	/* when modifying/adding a value */
	/* we are/will split where the value was */
	/* we then write out new string OR put out */
	/* a null key/value pair */
	if(value)
		sprintf(lpstr,"%s=%s\n",key,value);
	else
		sprintf(lpstr,"\n");

	np->szentry = lpstr;		/* new string (allocated) */
	np->offset  = 0;		/* next read is from beginning */
	np->next    = sp->next;		/* what is after this one */
	np->sname   = WinStrdup(sp->sname);  /* section name */
	np->entrysz = strlen(lpstr);	  /* section size, from datum */

	/* this entry now points to new entry */
	sp->next    = np;

	/* add section after the piece we broke off */
	if(flag) {
			tp = (SEGMENT *) GetFreeObject(SEGSIZE);
			tp->sflag = 0;
			tp->szentry = sp->offset;
			tp->offset  = 0;
			tp->sname = WinStrdup(sp->sname);
			tp->entrysz = sp->entrysz - (sp->offset - sp->szentry);

			tp->next = np->next;
			np->next = tp;
		
			/* this entry ends where this one left off */
			sp->entrysz = sp->szlast - sp->szentry;
		    sp->offset  = 0;	/* we will read from start next */
	}

	return np;
}

BOOL    WINAPI 
WritePrivateProfileString(LPCSTR section, LPCSTR key, LPCSTR value, LPCSTR file)
{
	SEGMENT *sp,*fp,*cp;
	char datum[512];
	char *lp,*vp;

	
    	APISTR((LF_APICALL, "WritePrivateProfileString(%p,%p,%p,%p)\n",
		section,key,value,file));
	
	/* as per the definition, flush the cache */
	if(section == NULL && key == NULL && value == NULL) {
    		APISTR((LF_APIRET, "WritePrivateProfileString flush cache(%s)\n",file));
		return FlushIniCache(file);
	}

	sp = LoadIniFile(file,1);
	fp = sp;

	/* for section header, search for section */
	while (sp) {

		if(strnicmp(sp->sname+1,section,strlen(section)) == 0) {
			/* if key is null, then we are deleting section */
			/* so point fp->next (previous section) to sp->next, */
			/* which will delete that section, and then */
			/* set sp to sp->next and continue with next section */
			/* for this case, do not update fp... */
			/* (in the event there are multiple sections with the */
			/* same name, this will delete those as well */

			/* found the section, now find the key... */
			sp->offset = 0;	/* guarantee to read from beginning */
			cp = sp;
			while((lp = getlines(&sp,section,datum,512))) {
				cp = sp;						/* the current section pointer */
				lp = strchr(datum,'=');		/* find terminating = sign */
				if (lp) {
					*lp = 0;   /* terminate the string */
					lp++;	   /* pointing past = sign */
						
					/* if we have blanks skip them */
					while(*lp == ' ')
						lp++;

					vp = lp;	/* we have a value */
					lp = datum;	/* this is the key */
					
					if(stricmp(key,lp) == 0) {
						/* found key, split section here... */
						addsection(sp,1,key,value);			
    						APISTR((LF_APIRET, "WritePrivateProfileString returns 1\n"));
						return 1;
					}
				}
				
				/* have not found key yet... */
				continue;  /* key did not match continue */
				
			}

			/* cp will have new value section linked to it */
			/* NOTE: since key was not found, we add value */
			/* iff value is not null */
			if(cp && value) {
				/* not split, just tacked on... */
				addsection(cp,0,key,value);			
    				APISTR((LF_APIRET, "WritePrivateProfileString returns 1\n"));
				return 1;
			}
    			APISTR((LF_APIRET, "WritePrivateProfileString returns 0\n"));
			return 0;
		}

		/* try next section */
		if(sp) {
			fp = sp;
			sp = sp->next;
		}
	}

/*
 * ChangeLog
 * added Jun 7, 1998
 * was not adding section, if it didn't already exist
 * now adds to end of file...
 */
	/* section not found, so add a new section, after all others */
	wsprintf(datum,"[%s]",section);
	cp = TWIN_GetNewObject(0,datum);
	
	/* fp was last valid section */
	/* could have been first unamed section */
	fp->next = cp;
	/* now add data behind this section */
	addsection(cp,0,key,value);			

    	APISTR((LF_APIRET, "WritePrivateProfileString returns 1\n"));
	return 1;
}

BOOL    WINAPI 
WriteProfileString(LPCSTR section, LPCSTR key, LPCSTR value)
{
	return WritePrivateProfileString(section,key,value,0);
}

DWORD	WINAPI 
GetPrivateProfileSectionNames(LPTSTR lpszReturnBuffer, DWORD nSize, LPCTSTR lpFileName)
{
	SEGMENT *sp;
	char *bp;
	char *sname;
	int   n,cnt;
	char *lp;
	
	/* find section header for file */
	sp = LoadIniFile(lpFileName,0);

	sname = "";
	bp = lpszReturnBuffer;			/* where to write in buffer */
	cnt = nSize;				/* remaining space in buffer */
	bp[0] = 0;				/* build in double terminator */
	bp[1] = 0;

	/* for section header, search for section */
	while (sp) {
		if(sp->sname[0] == '[' && strcmp(sname,sp->sname)) {

			/* since multiple sections can have same name */
			sname = sp->sname;

			strcpy(bp,sname+1);		/* copy it */
			if((lp = strchr(bp,']'))) {	/* try to ignore the ] character */
				*lp = 0;		/* ignore trailing garbage */
			}

			n = strlen(bp);		/* we added characters to bp, how many? */
			bp += n;		/* update pointer */
			cnt -= n;		/* decrement remaining space */
			*++bp = 0;		/* advance to next location */
			cnt--;			/* additional terminator */
		}
		
		/* try next section */
		sp = sp->next;
	}

	return 0;
}

/**********************************************************************/

#ifdef TWIN32

BOOL WINAPI
WritePrivateProfileSection(LPCTSTR lpszSection, LPCTSTR lpszKeysValues,
	LPCTSTR lpszFileName)
{
	char key[BUFSIZ];
	char *value;
	char *lpstr = (char *) lpszKeysValues;

	LOGSTR((LF_LOG,"WritePrivateProfileSection: filename %s\n",
		lpszFileName));

	for ( ; *lpstr != '\0'; lpstr += strlen(lpstr) + 1)
	{
		strncpy(key, lpstr, sizeof(key));
		if ((value = strchr(key, '=')) == NULL)
			break;
		*value = '\0';		/* terminate key with '\0' */
		value += 1;
		if (!WritePrivateProfileString(lpszSection, key, value,
			lpszFileName))
			return (FALSE);
	}
	return (TRUE);

}

BOOL WINAPI
WriteProfileSection(LPCTSTR lpszSection, LPCTSTR lpszKeysValues)
{
	return WritePrivateProfileSection(lpszSection, lpszKeysValues,0);
}

static char hex_digit[] = "0123456789abcdef";

static int decode_hex(int c)
{
	c = toupper(c);
	c -= '0';
	if (c > 9) c -= 7;
	return (c);
}

BOOL WINAPI
GetPrivateProfileStruct(LPCTSTR lpszSection,
	LPCTSTR lpszKey,
	LPVOID lpStruct,
	UINT uSizeStruct,
	LPCTSTR szFile)
{
	char valuebuf[BUFSIZ];
	char *src, *dst;

	LOGSTR((LF_LOG,"GetPrivateProfileStruct: filename %s\n",
		szFile));

	if (!GetPrivateProfileString(lpszSection, lpszKey, "", valuebuf,
		sizeof(valuebuf), szFile))
		return (FALSE);

	/* convert hex string to struct bytes */
	if (strlen(valuebuf) / 2 > uSizeStruct)
		return (FALSE);
	src = valuebuf;
	dst = (char *) lpStruct;
	for ( ; *src != '\0'; src += 2, dst++)
		*dst = decode_hex(src[0]) << 4 | decode_hex(src[1]);

	return (TRUE);

}

BOOL WINAPI
WritePrivateProfileStruct(LPCTSTR lpszSection,
	LPCTSTR lpszKey,
	LPVOID lpStruct,
	UINT uSizeStruct,
	LPCTSTR szFile)
{
	char valuebuf[BUFSIZ];
	char *value, *src, *dst;

	LOGSTR((LF_LOG,"WritePrivateProfileStruct: filename %s\n",
		szFile));

	if (lpStruct == NULL)
		value = NULL;
	else
	{
		value = valuebuf;
		/* encode struct bytes as hex string */
		if (2 * uSizeStruct + 1 > sizeof(valuebuf))
			return (FALSE);
		src = (char *) lpStruct;
		dst = valuebuf;
		for ( ; uSizeStruct-- > 0; src++, dst += 2)
		{
			dst[0] = hex_digit[(*src & 0xf0) >> 4];
			dst[1] = hex_digit[*src & 0x0f];
		}
		*dst = '\0';
	}

	return WritePrivateProfileString((LPSTR)lpszSection, (LPSTR)lpszKey, value, (LPSTR)szFile);

}

#endif /* twin32 */

/***************************************************************/

static SEGMENT *eventFree;
static SEGMENT *eventList;
static int    	eventCount;

/***************************************************************/

static SEGMENT *
GetFreeObject(int objsize)
{
	SEGMENT *next;
	SEGMENT *list;

	/* if we have a free one */
	if(eventFree) {
		next = eventFree;
		eventFree = (void *) next->next;
		memset((char *) next,0,objsize);
		return next;
	}

	/* prime the pump with another set of events */
	if(eventCount == 0) {
		eventCount = 32;
		eventList = (SEGMENT *) WinMalloc(eventCount*objsize);
	}
	
	/* point to the next available */
	eventCount--;	
	list = (SEGMENT *) eventList;
	next = &list[eventCount];
	memset((char *) next,0,objsize);

	return next;
}
